//
// Copyright (c) 2002-2021 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#ifndef __WZCPP_CONCURRENT_QUEUE_H__
#define __WZCPP_CONCURRENT_QUEUE_H__

#include "wzcpp/features/utils/Memory.h"
#include <chrono>
#include <condition_variable>
#include <memory>
#include <mutex>
#include <queue>


namespace wzcpp
{
/*!
 * This is a threadsafe wrapper around the std::queue container
 */
template<typename T>
class ConcurrentQueue
{
private:
	mutable std::mutex mut;
	std::queue<T> data_queue;
	std::condition_variable data_cond;

public:
	ConcurrentQueue() = default;

	/*!
	Copy constructor
	*/
	ConcurrentQueue(ConcurrentQueue const& other)
	{
		std::lock_guard<std::mutex> lock(other.mut);
		data_queue = other.data_queue;
	}

	/*!
	 * Add an value at the end of the queue
	 */
	void push(T&& new_value)
	{
		{
			std::lock_guard<std::mutex> lock(mut);
			data_queue.push(std::move(new_value));
		}
		data_cond.notify_one();
	}

	/*!
	 * Add an value at the end of the queue
	 */
	void push(const T& new_value)
	{
		{
			std::lock_guard<std::mutex> lock(mut);
			data_queue.push(new_value);
		}
		data_cond.notify_one();
	}

	template<class... Args>
	void emplace(Args&&... args)
	{
		{
			std::lock_guard<std::mutex> lock(mut);
			data_queue.emplace(std::forward<Args>(args)...);
		}
		data_cond.notify_one();
	}

	/*!
	 * Wait infinitly for new data in the queue, if data is ready retrieve the first
	 * element into the given reference value
	 */
	void wait_and_pop(T& value)
	{
		std::unique_lock<std::mutex> lock(mut);
		data_cond.wait(lock, [this] { return !data_queue.empty(); });
		value = data_queue.front();
		data_queue.pop();
	}

	/*!
	 * Wait infinitly for new data in the queue, if data is ready retrieve the first
	 * element into a newly created unique pointer
	 */
	std::unique_ptr<T> wait_and_pop()
	{
		std::unique_lock<std::mutex> lock(mut);
		data_cond.wait(lock, [this] { return !data_queue.empty(); });
		auto res(wzcpp::make_unique<T>(data_queue.front()));
		data_queue.pop();
		return res;
	}

	/*!
	 * Wait for a given duration for new data in the queue, if data is ready retrieve the
	 * first element into the given reference value
	 */
	template<class Rep, class Period>
	std::cv_status wait_for_and_pop(T& value, const std::chrono::duration<Rep, Period>& rel_time)
	{
		std::unique_lock<std::mutex> lock(mut);
		std::cv_status status = std::cv_status::timeout;
		bool result           = data_cond.wait_for(lock, rel_time, [this] { return !data_queue.empty(); });
		if (result)
		{
			value = std::move(data_queue.front());
			data_queue.pop();
			status = std::cv_status::no_timeout;
		}
		return status;
	}

	/*!
	 * Wait until a given timepoint is reached or new data is available in the queue, if
	 * data is ready retrieve the first element into the given reference value
	 */
	template<typename Clock>
	std::cv_status wait_until_and_pop(T& value, const std::chrono::time_point<Clock>& timeout_time)
	{
		std::unique_lock<std::mutex> lock(mut);
		std::cv_status status = std::cv_status::timeout;
		bool result = data_cond.wait_until(lock, timeout_time, [this] { return !data_queue.empty(); });
		if (result)
		{
			value = std::move(data_queue.front());
			data_queue.pop();
			status = std::cv_status::no_timeout;
		}
		return status;
	}

	/*!
	 * Wait for a given duration for new data in the queue, if data is ready retrieve the
	 * first element into a newly created unique pointer, if a timeout occurs the returned
	 * unique_ptr is empty
	 */
	template<class Rep, class Period>
	std::unique_ptr<T> wait_for_and_pop(const std::chrono::duration<Rep, Period>& rel_time)
	{
		std::unique_lock<std::mutex> lock(mut);
		bool result = data_cond.wait_for(lock, rel_time, [this] { return !data_queue.empty(); });
		if (result)
		{
			std::unique_ptr<T> res(wzcpp::make_unique<T>(std::move(data_queue.front())));
			data_queue.pop();
			return res;
		}
		return {};
	}

	/*!
	 * Wait until a given timepoint is reached or new data is available in the queue, if
	 * data is ready retrieve the first element into a newly created unique pointer, if a
	 * timeout occurs the returned unique_ptr is empty
	 */
	template<class Clock>
	std::unique_ptr<T> wait_until_and_pop(const std::chrono::time_point<Clock>& timeout_time)
	{
		std::unique_lock<std::mutex> lock(mut);
		bool status = data_cond.wait_until(lock, timeout_time, [this] { return !data_queue.empty(); });
		if (status)
		{
			std::unique_ptr<T> res(wzcpp::make_unique<T>(data_queue.front()));
			data_queue.pop();
			return res;
		}
		return {};
	}

	/*!
	 * Sets the given reference to the first value in quque if there are any elements in
	 * queue.
	 * @return true if the queue contained an element to return, false otherwise
	 */
	bool try_pop(T& value)
	{
		std::lock_guard<std::mutex> lock(mut);
		if (data_queue.empty()) return false;
		value = std::move(data_queue.front());
		data_queue.pop();
		return true;
	}

	/*!
	 * If data is ready retrieve the first element
	 * into a newly created unique pointer
	 */
	std::unique_ptr<T> try_pop()
	{
		std::lock_guard<std::mutex> lock(mut);
		if (data_queue.empty()) return std::unique_ptr<T>();
		std::unique_ptr<T> res(wzcpp::make_unique<T>(data_queue.front()));
		data_queue.pop();
		return res;
	}

	bool empty() const
	{
		std::lock_guard<std::mutex> lock(mut);
		return data_queue.empty();
	}

	std::size_t size() const
	{
		std::lock_guard<std::mutex> lock(mut);
		return data_queue.size();
	}
};


/*!************************************************************
 * This is a threadsafe wrapper around the std::deque container
 */
template<typename T>
class ConcurrentDeque
{
private:
	mutable std::mutex mut;
	std::deque<T> data_queue;
	std::condition_variable data_cond;

public:
	ConcurrentDeque() = default;

	/*!
	Copy constructor
	*/
	ConcurrentDeque(ConcurrentDeque const& other)
	{
		std::lock_guard<std::mutex> lock(other.mut);
		data_queue = other.data_queue;
	}

	/*!
	 * Add an value at the end of the queue
	 */
	void push_back(T&& new_value)
	{
		{
			std::lock_guard<std::mutex> lock(mut);
			data_queue.push_back(std::move(new_value));
		}
		data_cond.notify_one();
	}

	/*!
	 * Add an value at the end of the queue
	 */
	void push_back(const T& new_value)
	{
		{
			std::lock_guard<std::mutex> lock(mut);
			data_queue.push_back(new_value);
		}
		data_cond.notify_one();
	}

	template<class... Args>
	void emplace_back(Args&&... args)
	{
		{
			std::lock_guard<std::mutex> lock(mut);
			data_queue.emplace_back(std::forward<Args>(args)...);
		}
		data_cond.notify_one();
	}

	/*!
	 * Add an value at the front of the queue
	 */
	void push_front(T&& new_value)
	{
		{
			std::lock_guard<std::mutex> lock(mut);
			data_queue.push_front(std::move(new_value));
		}
		data_cond.notify_one();
	}

	/*!
	 * Add an value at the front of the queue
	 */
	void push_front(const T& new_value)
	{
		{
			std::lock_guard<std::mutex> lock(mut);
			data_queue.push_front(new_value);
		}
		data_cond.notify_one();
	}

	template<class... Args>
	void emplace_front(Args&&... args)
	{
		{
			std::lock_guard<std::mutex> lock(mut);
			data_queue.emplace_front(std::forward<Args>(args)...);
		}
		data_cond.notify_one();
	}

	/*!
	 * Wait infinitly for new data in the queue, if data is ready retrieve the first
	 * element into the given reference value
	 */
	void wait_and_pop_front(T& value)
	{
		std::unique_lock<std::mutex> lock(mut);
		data_cond.wait(lock, [this] { return !data_queue.empty(); });
		value = std::move(data_queue.front());
		data_queue.pop_front();
	}

	/*!
	 * Wait infinitly for new data in the queue, if data is ready retrieve the first
	 * element into a newly created unique pointer
	 */
	std::unique_ptr<T> wait_and_pop_front()
	{
		std::unique_lock<std::mutex> lock(mut);
		data_cond.wait(lock, [this] { return !data_queue.empty(); });
		std::unique_ptr<T> res(wzcpp::make_unique<T>(data_queue.front()));
		data_queue.pop();
		return res;
	}

	/*!
	 * Wait for a given duration for new data in the queue, if data is ready retrieve the
	 * first element into the given reference value
	 */
	template<class Rep, class Period>
	std::cv_status wait_for_and_pop_front(T& value, const std::chrono::duration<Rep, Period>& rel_time)
	{
		std::unique_lock<std::mutex> lock(mut);
		std::cv_status status = std::cv_status::timeout;
		bool result           = data_cond.wait_for(lock, rel_time, [this] { return !data_queue.empty(); });
		if (result)
		{
			value = std::move(data_queue.front());
			data_queue.pop_front();
			status = std::cv_status::no_timeout;
		}
		return status;
	}

	/*!
	 * Wait until a given timepoint is reached or new data is available in the queue, if
	 * data is ready retrieve the first element into the given reference value
	 */
	template<typename Clock>
	std::cv_status wait_until_and_pop_front(T& value, const std::chrono::time_point<Clock>& timeout_time)
	{
		std::unique_lock<std::mutex> lock(mut);
		std::cv_status status = std::cv_status::timeout;
		bool result = data_cond.wait_until(lock, timeout_time, [this] { return !data_queue.empty(); });
		if (result)
		{
			value = std::move(data_queue.front());
			data_queue.pop_front();
			status = std::cv_status::no_timeout;
		}
		return status;
	}

	/*!
	 * Wait for a given duration for new data in the queue, if data is ready retrieve the
	 * first element into a newly created unique pointer, if a timeout occurs the returned
	 * unique_ptr is empty
	 */
	template<class Rep, class Period>
	std::unique_ptr<T> wait_for_and_pop_front(const std::chrono::duration<Rep, Period>& rel_time)
	{
		std::unique_lock<std::mutex> lock(mut);
		bool result = data_cond.wait_for(lock, rel_time, [this] { return !data_queue.empty(); });
		if (result)
		{
			std::unique_ptr<T> res(wzcpp::make_unique<T>(std::move(data_queue.front())));
			data_queue.pop_front();
			return res;
		}
		return {};
	}

	/*!
	 * Wait until a given timepoint is reached or new data is available in the queue, if
	 * data is ready retrieve the first element into a newly created unique pointer, if a
	 * timeout occurs the returned unique is empty
	 */
	template<class Clock>
	std::unique_ptr<T> wait_until_and_pop_front(const std::chrono::time_point<Clock>& timeout_time)
	{
		std::unique_lock<std::mutex> lock(mut);
		bool status = data_cond.wait_until(lock, timeout_time, [this] { return !data_queue.empty(); });
		if (status)
		{
			std::unique_ptr<T> res(wzcpp::make_unique<T>(data_queue.front()));
			data_queue.pop_front();
			return res;
		}
		return {};
	}

	/*!
	 * Sets the given reference to the first value in quque if there are any elements in
	 * queue.
	 * @return true if the queue contained an element to return, false otherwise
	 */
	bool try_pop_front(T& value)
	{
		std::lock_guard<std::mutex> lock(mut);
		if (data_queue.empty()) return false;
		value = std::move(data_queue.front());
		data_queue.pop_front();
		return true;
	}

	/*!
	 * If data is ready retrieve the first element
	 * into a newly created unique pointer
	 */
	std::unique_ptr<T> try_pop_front()
	{
		std::lock_guard<std::mutex> lock(mut);
		if (data_queue.empty()) return std::unique_ptr<T>();
		std::unique_ptr<T> res(wzcpp::make_unique<T>(data_queue.front()));
		data_queue.pop_front();
		return res;
	}

	bool empty() const
	{
		std::lock_guard<std::mutex> lock(mut);
		return data_queue.empty();
	}

	std::size_t size() const
	{
		std::lock_guard<std::mutex> lock(mut);
		return data_queue.size();
	}

	void clear()
	{
		std::lock_guard<std::mutex> lock(mut);
		data_queue.clear();
	}

	void shrink_to_fit()
	{
		std::lock_guard<std::mutex> lock(mut);
		data_queue.shrink_to_fit();
	}

	void resize(std::size_t count)
	{
		std::lock_guard<std::mutex> lock(mut);
		data_queue.resize(count);
	}

	void resize(std::size_t count, T value = T())
	{
		std::lock_guard<std::mutex> lock(mut);
		data_queue.resize(count, value);
	}


	void resize(std::size_t count, const T& value)
	{
		std::lock_guard<std::mutex> lock(mut);
		data_queue.resize(count, value);
	}
};
}  // end namespace wzcpp
#endif
