//
// Copyright (c) 2021 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#ifndef __WZCPP_TYPE_CONVERTER_H__
#define __WZCPP_TYPE_CONVERTER_H__

#include "wzcpp/features/utils/Formatter.h"

#include <type_traits>
#include <cstdint>
#include <tuple>
#include <array>
#include <stdexcept>

namespace wzcpp
{

	template <typename T, typename = int>
    struct has_resize : std::false_type { };

    template <typename T>
    struct has_resize <T, decltype( (void) std::declval<T>().resize(1), 0)> : std::true_type { };
    
    enum class BitFlags
    {
        BIT00 = (1 << 0), // Bit 0
        BIT01 = (1 << 1), // Bit 1
        BIT02 = (1 << 2), // Bit 2
        BIT03 = (1 << 3), // Bit 3
        BIT04 = (1 << 4), // Bit 4
        BIT05 = (1 << 5), // Bit 5
        BIT06 = (1 << 6), // Bit 6
        BIT07 = (1 << 7)  // Bit 7
    };

    /*!
    @brief Get the high byte of a value

    @tparam T The type from which the high byte should be retrieved.
    @param value The actual value to extract from
    @return std::uint8_t  The high byte or the second byte if T has more than 2 bytes
    */
    template<typename T>
    std::uint8_t highByte(T value)
    {
        assert(sizeof(T) > 1);
        return static_cast<std::uint8_t>((value >> 8) & 0xff);
    };

    /*!
        @brief Get the low byte of a value

        @tparam T The type from which the low byte should be retrieved.
        @param value The actual value to extract from
        @return std::uint8_t The low byte
    */
    template<typename T>
    std::uint8_t lowByte(const T value)
    {
        assert(sizeof(T) > 1);
        return static_cast<std::uint8_t>(value & 0xff);
    };

    /*!
    @brief Converts a iterable list of bytes to a given integral ReturnType

    @tparam BytesTypeIter An iterator type
    @tparam ReturnType -The integral type in which we want to convert the bytes
    @param bytesIter An iterator pointing to the first byte of the bytes to convert ( Make sure that there a at least
    sizeof(ReturnType)) bytes after the iterator
    @return ReturnType Returns the converted value
    */
    template<class BytesTypeIter, class ReturnType, std::enable_if_t<std::is_integral<ReturnType>::value>* = nullptr>
    ReturnType convertBytesTo(BytesTypeIter bytesIter)
    {
        ReturnType result{0};
        constexpr auto returnTypeSize{sizeof(ReturnType)};

        for (std::size_t index = 1; index <= returnTypeSize; index++)
        {
            result |= static_cast<ReturnType>(*bytesIter) << (8 * (returnTypeSize - index));
            bytesIter++;
        }
        return result;
    }

    /*!
    @brief Converts a iterable list of bytes to a given enum ReturnType

    @tparam BytesTypeIter An iterator type
    @tparam ReturnType The enum type in which we want to convert the bytes
    @param bytesIter An iterator pointing to the first byte of the bytes to convert ( Make sure that there a at least
    sizeof(std::underlying_type_t<ReturnType>)) bytes after the iterator
    @return ReturnType Returns the converted value
    */
    template<class BytesTypeIter, class ReturnType, std::enable_if_t<std::is_enum<ReturnType>::value>* = nullptr>
    ReturnType convertBytesTo(BytesTypeIter bytesIter)
    {
        using underlying_type = std::underlying_type_t<ReturnType>;

        underlying_type underlyingResult =
            convertBytesTo<BytesTypeIter, underlying_type>(std::forward<BytesTypeIter>(bytesIter));

        return static_cast<ReturnType>(underlyingResult);
    }

    /*!
    @brief Converts a given integral type into a list of bytes

    @tparam BytesType The container type for the bytes to store in
    @tparam InputType The integral variable type which should be converted
    @param input The input variable to convert
    @return BytesType Returns a container of type BytesType filled with the bytes from the input parameter
    */
    template<class BytesType, class InputType, std::enable_if_t<std::is_integral<InputType>::value>* = nullptr>
    BytesType convertBytesFrom(InputType input)
    {
        constexpr auto inputTypeSize{sizeof(InputType)};
        BytesType result;
        std::fill(std::begin(result), std::end(result), 0);        
        
        if constexpr(has_resize<BytesType>::value)
        {
            result.resize(inputTypeSize);
        }

        if(inputTypeSize > result.size())
        {
            std::ostringstream oss;
            oss << "Incompatible sizes " << inputTypeSize << " (input) -> " << result.size() << " (output)";
            throw std::runtime_error(oss.str());
        }
        const auto offset = result.size() - inputTypeSize;
        for (std::size_t index = 1; index <= inputTypeSize; index++)
        {
            result[index-1+offset]=static_cast<typename BytesType::value_type>(input >> (8 * (inputTypeSize - index)));
        }
        return result;
    }


    /*!
    @brief Converts a given enum type into a list of bytes

    @tparam BytesType The container type for the bytes to store in
    @tparam InputType The enum variable type which should be converted
    @param input The input variable to convert
    @return BytesType Returns a container of type BytesType filled with the bytes from the input parameter
    */
    template<class BytesType, class InputType, std::enable_if_t<std::is_enum<InputType>::value>* = nullptr>
    BytesType convertBytesFrom(InputType input)
    {
        using underlying_type = std::underlying_type_t<InputType>;
        underlying_type underlyingInput = static_cast<underlying_type>(input);


        return convertBytesFrom<BytesType, underlying_type>(underlyingInput);
    }
}  // namespace wzcpp
#endif  // __WZCPP_TYPE_CONVERTER_H__
