//
// Copyright (c) 2021 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#ifndef __WZCPP_KNX_FLOATCONVERTER_H__
#define __WZCPP_KNX_FLOATCONVERTER_H__

#include "wzcpp/features/converter/type_conv.h"
#include "wzcpp/features/error/Exception.h"

#include <array>
#include <stdexcept>

namespace wzcpp
{
    enum class KnxFloatErrorCodes
    {
        Float16ConversionError = 100,
    };

    DECLARE_WZ_EXCEPTION(
        Float16ConversionException,
        KnxFloatErrorCodes::Float16ConversionError,
        wzcpp::error::WzBaseException);  // Exception class that can be used if a function is not implemented and should throw an error
   
    using Knx2ByteFloat = std::array<std::uint8_t,2>;


    class Knx2ByteFloatConverter
    {
        public:
            /*!
                converts a 4-byte float to a 2-byte float
                This function converts a 4-byte float (IEEE-Float) to
                a 2-byte float (KNX-Float DPT 9.xxx/EIS5)
            */
            static Knx2ByteFloat fromFloat4(float f)
            {
                Knx2ByteFloat knxFloat{0,0};
                if (!KnxDpt_ConvIeeeFlToKnxFl(f, knxFloat))
                {
                    throw Float16ConversionException(WZ_SL, "Conversion from float32 failed");
                }
                return knxFloat;
            }
            /*!
                This function converts a 2-byte float (KNX-Float DPT 9.xxx/EIS5) to a 4-byte float (IEEE-Float)
                The format of the KNX-float is MEEEEMMM MMMMMMMM
            */
            static float toFloat4(Knx2ByteFloat& knxFloat)
            {                
                float ieeeFloat = 0;

                if (!KnxDpt_ConvKnxFlToIeeeFl(knxFloat, ieeeFloat))
                {
                    throw Float16ConversionException(WZ_SL, "Conversion to float32 failed");
                }

                return ieeeFloat;  
            }
            /*!
                This function converts a 2-byte float (KNX-Float DPT 9.xxx/EIS5) to a 4-byte float (IEEE-Float)
                The format of the KNX-float is MEEEEMMM MMMMMMMM
            */
            static float toFloat4(Knx2ByteFloat&& knxFloat)
            {                
                float ieeeFloat = 0;

                if (!KnxDpt_ConvKnxFlToIeeeFl(knxFloat, ieeeFloat))
                {
                    throw Float16ConversionException(WZ_SL, "Conversion to float32 failed");
                }

                return ieeeFloat;  
            }
        private:
            static bool KnxDpt_ConvIeeeFlToKnxFl(float fIeeeFloat, Knx2ByteFloat& knxFloat)
            {
                std::int32_t nInt; // Value * 100

                fIeeeFloat *= 100.0; // Map value to float*100
                if (fIeeeFloat > 0)  // If value positive
                {
                    nInt = (std::int32_t)(fIeeeFloat + 0.5); // Map to signed int*100
                }
                else                              // Else: Value negative
                {
                    nInt = (std::int32_t)(fIeeeFloat - 0.5); // Map to signed int*100
                }

                return KnxDpt_ConvSInt32ToKnxFl(nInt, knxFloat); // Do the conversion
            }
            
            /// This function converts a long with scaling * 100 to
            /// a KNX-Float (DPT 9.xxx/EIS5)
            /// Return: true if conversion was successful
            /// 		false on error
            static bool KnxDpt_ConvSInt32ToKnxFl(int nInt, Knx2ByteFloat& knxFloat)
                {
                    std::uint32_t nMantissa;      // Mantissa of Knx-Float
                    std::uint16_t nMantissaMax; // Maxima of mantissa
                    std::uint8_t nExponent;     // Exponent of Knx-Float
                    std::uint8_t nLostBit;      // Last shifted bit = lost bit

                    if (nInt >= 0) // If value positive
                    {
                        nMantissa = (std::uint32_t)nInt; // Init mantissa
                        nMantissaMax = 2047;            // Maxima of pos range
                    }
                    else                              // Else: Value negative
                    {
                        nMantissa = (std::uint32_t)-nInt; // Init positive mantissa
                        nMantissaMax = 2048;             // Maxima of neg range
                    }

                    // Note:
                    //		If we shift nMantissa right, and nMantissa is odd we lost a bit!
                    //		For the last lost bit, we increase nMantissa, this is an
                    //		implicit binary rounding up of nMantissa.
                    //		"(nMantissa + nLostBit)" in the while loop, make sure
                    //		that we have no overflow in nMantissa.
                    nLostBit = 0;  // Set lost bit to 0
                    nExponent = 0; // Set exponent to 0
                    while ((nMantissa + nLostBit) >
                        nMantissaMax) // Loop until (nInt / 2^nExponent) <= nMantissaMax
                    {
                        nLostBit = (std::uint8_t)(nMantissa & 0x01); // Remember the lost bit in
                        // the next shift operation
                        nMantissa >>= 1;                              // Calc 2^nExponent
                        nExponent++;                                  // Calc nExponent
                    }

                    nMantissa += nLostBit; // Round up nMantissa

                    if (nExponent > 15) // If value is out of range
                    {
                        // Error: Value out of KNX float16 range
                        knxFloat[0] = 0x7F; // Set Knx float to error value
                        knxFloat[1] = 0xFF; // Set Knx float to error value
                        return false;         // Return error
                    }

                    if (nInt >= 0) // If value positive
                    {
                        knxFloat[0] = 0x00; // Set the sign to pos
                    }
                    else                    // Else: Value negative
                    {
                        nMantissa = 2048 - nMantissa; // Calculate a signed mantissa
                        knxFloat[0] = static_cast<std::uint8_t>(BitFlags::BIT07);        // Set the sign to neg
                    }

                    // Set value to format SEEEEMMM MMMMMMMM
                    // Sign already set before
                    knxFloat[0] |= (nExponent << 3) & 0x78;  // Set sEEEEmmm
                    knxFloat[0] |= highByte(nMantissa); // Set seeeMMM
                    knxFloat[1] = lowByte(nMantissa);  // Set MMMMMMMM

                    return true; // Return success
                }

                /// This function converts a KNX-Float (DPT 9.xxx/EIS5) to a IEEE-Float
                /// The format of the KNX-float is MEEEEMMM MMMMMMMM
                /// Return: true if conversion was successful
                /// 		false on error
                static bool KnxDpt_ConvKnxFlToIeeeFl(const Knx2ByteFloat& knxFloat, float& fValue)
                {
                    int nLongInt; // Value*100 as int

                    if (KnxDpt_ConvKnxFlToSInt32(knxFloat,
                                                nLongInt)) // If conversation was success
                    {
                        fValue = static_cast<float>(nLongInt / 100.0); // Adapt value range
                        return true;                                     // Return success
                    }

                    return false; // Return error
                }
                

                /// This function converts a Knx-Float (DPT 9.xxx/EIS5)
                /// to a long with scaling * 100
                /// Return: true if conversion was successful
                /// 		false on error
                static bool KnxDpt_ConvKnxFlToSInt32(const Knx2ByteFloat& knxFloat, std::int32_t& nInt)
                {
                    std::uint16_t nRaw;      // KNX float raw data
                    std::uint16_t nMultipli; // Multiplicator
                    int nMantissa;            // Mantissa

                    if ((knxFloat[0] == 0x7F) &&
                        (knxFloat[1] == 0xFF)) // If Knx float error value
                    {
                        // invalid KNX value
                        return false; // Return error
                    }

                    nRaw = convertBytesTo<Knx2ByteFloat::const_iterator, std::uint16_t>(knxFloat.begin()); // Build the raw word
                    nMantissa = nRaw & 0x07FF;                       // Extract pos. mantissa
                    nMultipli = static_cast<std::uint8_t>(BitFlags::BIT00) << ((knxFloat[0] >> 3) & 0x0F); // Build 2^(exponent)

                    if (knxFloat[0] & 0x80) // If value negative
                    {
                        // Mathematic for neg signment:
                        //
                        //	FloatValue = (nMantissa - 2048) * 2^nExponent / 100.0
                        //
                        //	Mantissa is ranged from -2048 to 2047 and 100 times of value.
                        //	E.g. if you have only set the signed bit,
                        //	the result is (0-2048)/100 = - 20.48 and not 0.0
                        nMantissa -= 2048; // Build complement
                    }

                    nInt = (int)nMantissa * nMultipli; // Build signed int

                    return true; // Return success
                }

                    };

}  // namespace wzcpp
#endif  // __WZCPP_KNX_FLOATCONVERTER_H__
