//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#ifndef __WZCPP_LOGGING_LOGGER_INTERFACE_H__
#define __WZCPP_LOGGING_LOGGER_INTERFACE_H__

#include "detail.h"

#include "wzcpp/features/utils/Formatter.h"
#include "wzcpp/features/utils/String.h"
#include "wzcpp/features/utils/StringView.h"

#include <spdlog/sinks/daily_file_sink.h>
#include <spdlog/sinks/rotating_file_sink.h>
#include <spdlog/sinks/stdout_color_sinks.h>
#include <spdlog/spdlog.h>

#include <bitset>
#include <cstring>
#include <iomanip>
#include <ios>
#include <iostream>
#include <sstream>
#include <type_traits>

#if (USE_POCO_LEGACY_LOGGER == 1)
#include "Poco/Logger.h"
#endif


namespace spdlog
{
inline level::level_enum from_wzlevel(wzcpp::LogLevel lvl)
{
	switch (lvl)
	{
		case LVL_VERBOSE:
			return spdlog::level::trace;
		case LVL_DEBUG:
			return spdlog::level::debug;
		case LVL_INFO:
			return spdlog::level::info;
		case LVL_WARN:
			return spdlog::level::warn;
		case LVL_ERROR:
			return spdlog::level::err;
		case LVL_CRITICAL:
			return spdlog::level::critical;
		case wzcpp::LogLevel::fatal:
			return spdlog::level::critical;
		case wzcpp::LogLevel::notice:
			return spdlog::level::info;
		default:
			return spdlog::level::off;
	}
}
}  // namespace spdlog

namespace wzcpp
{
/***********************************************************************************************
 ** LoggerConfig
 **********************************************************************************************/
enum class LogTarget : int
{
	none                                 = 0,
	console                              = 1,
	rotatefile                           = 2,
	console_and_rotatefile               = 3,
	dailyfile                            = 4,
	console_and_dailyfile                = 5,
	rotatefile_and_dailyfile             = 6,
	console_and_rotatefile_and_dailyfile = 7
};

struct LoggerConfig
{
	std::string name        = "";
	LogLevel level          = LVL_ERROR;
	LogTarget targets       = LogTarget::none;
	std::string logFilePath = ""; /*!< path for the rotating or daily log files */
	int maxSizeMb;                /*!< Size of log files in MiB */
	int maxRotateFiles;           /*!< Max number of rotating log files */
	int maxDailyFiles;            /*!< Max number of daily log files */
	int logHour;                  /*!< Hour to write the daily log file */
	int logMinute;                /*!< Hour to write the daily log file */
	bool truncateDaily;           /*!< If true overrides the daily file */
};

class LogStream2 : public std::ostringstream
{
public:
	LogStream2(LogLevel lvl) : logLevel_(lvl)
	{
	}

	LogStream2(LogStream2&& other) : std::ostringstream(std::move(other))
	{
		set_rdbuf(other.rdbuf());
		logLevel_ = std::move(other.logLevel_);
	}

	LogLevel getLevel()
	{
		return logLevel_;
	}

	void setBuffer(std::streambuf* buffer)
	{
		set_rdbuf(buffer);
	}

	friend LogStream2& operator<<(LogStream2& lhs, const std::ostringstream& rhs)
	{
		lhs << rhs.str();
		return lhs;
	}

private:
	LogLevel logLevel_;
};


/***********************************************************************************************
 ** Basic loggerclass
 **********************************************************************************************/
/*!
    Base class for defining functions every
    logger should implement.
 */
#if USE_POCO_LEGACY_LOGGER == 1
using ChannelInterface = Poco::Channel*;
#else
using ChannelInterface = std::shared_ptr<spdlog::sinks::sink>;
#endif

class BaseLogger
{

public:
	BaseLogger()
	{
		init();
	}

	BaseLogger(const std::string& name, LogLevel level) : name_(name), logLevel_(level)
	{
		init();
	}

	BaseLogger(BaseLogger& other) : name_(other.name_), logLevel_(other.logLevel_)
	{
	}

	BaseLogger(BaseLogger&& other) : name_(std::move(other.name_)), logLevel_(std::move(other.logLevel_))
	{
	}

	~BaseLogger() = default;

	const std::string& getName() const
	{
		return name_;
	}

	std::string& getName()
	{
		return name_;
	}

	void setName(const std::string& s)
	{
		name_ = s;
	}

	/*! Returns the least loggable level of the logger. */
	virtual LogLevel getLevel()
	{
		return logLevel_;
	}

	/*! Sets the least loggable level of the logger. */
	virtual void setLevel(LogLevel lvl)
	{
		logLevel_ = lvl;
	}

	/*!
	    @brief Initializes the Logger.
	    The base implementation is empty and needs to be
	    overridden by the implementing class.
	*/
	virtual void init()
	{
	}

	virtual void verbose(const std::string&)            = 0;
	virtual void debug(const std::string&)              = 0;
	virtual void info(const std::string&)               = 0;
	virtual void warn(const std::string&)               = 0;
	virtual void error(const std::string&)              = 0;
	virtual void critical(const std::string&)           = 0;
	virtual void addLoggingChannel(ChannelInterface)    = 0;
	virtual void removeLoggingChannel(ChannelInterface) = 0;

	virtual void stream(LogLevel lvl, std::ostringstream& stream)
	{
		switch (lvl)
		{
			case LogLevel::off:
				break;
			case LogLevel::verbose:
			case LogLevel::notice:
				verbose(stream.str());
				break;
			case LogLevel::debug:
				debug(stream.str());
				break;
			case LogLevel::info:
				info(stream.str());
				break;
			case LogLevel::warn:
				warn(stream.str());
				break;
			case LogLevel::error:
				error(stream.str());
				break;
			case LogLevel::critical:
			case LogLevel::fatal:
				critical(stream.str());
				break;
		}
	}

protected:
	std::string name_;
	/*! Least loggable level of the logger. */
	LogLevel logLevel_;
};

/*!
    Creates loggers for each configuration that is contained in the passed vector.
    All created log targets are multithreaded and threadsafe.
    If a configuration does not set the logFilePath then the path is
    set to ./[loggername].{daily}.log"
 */
inline void setConfiguration(std::vector<LoggerConfig> loggerConfigurations)
{
	for (auto& config : loggerConfigurations)
	{
		std::vector<spdlog::sink_ptr> sinks;
		std::bitset<4> bs(static_cast<int>(config.targets));
		auto bitString       = bs.to_string();
		bool logToDailyFile  = bitString[1] == '1';
		bool logToRotateFile = bitString[2] == '1';
		bool logToConsole    = bitString[3] == '1';

		if (logToConsole)
		{
			sinks.push_back(std::make_shared<spdlog::sinks::stdout_color_sink_mt>(spdlog::color_mode::automatic));
		}
		if (logToRotateFile)
		{
			sinks.push_back(std::make_shared<spdlog::sinks::rotating_file_sink_mt>(
			    config.logFilePath != "" ? config.logFilePath : config.name + ".log",
			    1024 * 1024 * config.maxSizeMb,
			    config.maxRotateFiles,
			    false));
		}
		if (logToDailyFile)
		{
			sinks.push_back(std::make_shared<spdlog::sinks::daily_file_sink_mt>(
			    config.logFilePath != "" ? config.logFilePath : config.name + ".daily.log",
			    config.logHour,
			    config.logMinute,
			    config.truncateDaily,
			    config.maxDailyFiles));
		}

		auto logger = std::make_shared<spdlog::logger>(config.name, std::begin(sinks), std::end(sinks));
		logger->set_level(spdlog::from_wzlevel(config.level));
		spdlog::register_logger(logger);
	}
}
}  // namespace wzcpp
#endif  // !__WZCPP_LOGGING_LOGGER_INTERFACE_H__
