//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#if (USE_POCO_LEGACY_LOGGER == 1)

#ifndef __WZCPP_LOGGING_POCOLEGACYLOGGER_H__
#define __WZCPP_LOGGING_POCOLEGACYLOGGER_H__

#include "LoggerInterface.h"
#include <Poco/AutoPtr.h>
#include <Poco/Channel.h>
#include <Poco/LogStream.h>
#include <Poco/Message.h>
#include <Poco/SplitterChannel.h>
#include <atomic>


namespace wzcpp
{

inline LogLevel fromPocoLogLevel(const Poco::Message::Priority level)
{
	switch (level)
	{
		case Poco::Message::PRIO_FATAL:
			return LogLevel::fatal;
		case Poco::Message::PRIO_CRITICAL:
			return LogLevel::critical;
		case Poco::Message::PRIO_WARNING:
			return LogLevel::warn;
		case Poco::Message::PRIO_ERROR:
			return LogLevel::error;
		case Poco::Message::PRIO_INFORMATION:
			return LogLevel::info;
		case Poco::Message::PRIO_NOTICE:
			return LogLevel::notice;
		case Poco::Message::PRIO_DEBUG:
			return LogLevel::debug;
		case Poco::Message::PRIO_TRACE:
			return LogLevel::verbose;
		default:
			return LogLevel::off;
	}
}

inline Poco::Message::Priority toPocoLogLevel(LogLevel level)
{

	switch (level)
	{
		case LogLevel::off:
			return static_cast<Poco::Message::Priority>(0);
		case LogLevel::fatal:
			return Poco::Message::PRIO_FATAL;
		case LogLevel::critical:
			return Poco::Message::PRIO_CRITICAL;
		case LogLevel::warn:
			return Poco::Message::PRIO_WARNING;
		case LogLevel::error:
			return Poco::Message::PRIO_ERROR;
		case LogLevel::info:
			return Poco::Message::PRIO_INFORMATION;
		case LogLevel::notice:
			return Poco::Message::PRIO_NOTICE;
		case LogLevel::debug:
			return Poco::Message::PRIO_DEBUG;
		case LogLevel::verbose:
			return Poco::Message::PRIO_TRACE;
		default:
			return static_cast<Poco::Message::Priority>(0);
	}
}

class PocoLegacyLogger : public BaseLogger
{
public:
	using BaseLogger::BaseLogger;

	PocoLegacyLogger() : BaseLogger()
	{
	}

	PocoLegacyLogger(const std::string& name, LogLevel level) : BaseLogger(name, level)
	{
		init();
	}

	static PocoLegacyLogger root()
	{
		PocoLegacyLogger rootLogger;
		auto& pocoRootLogger = Poco::Logger::root();
		rootLogger.setName(pocoRootLogger.name());
		rootLogger.setLevel(fromPocoLogLevel(static_cast<Poco::Message::Priority>(pocoRootLogger.getLevel())));
		return rootLogger;
	}

	void init() override
	{
		setLevel(logLevel_);
	}

	LogLevel getLevel() override
	{
		return fromPocoLogLevel(static_cast<Poco::Message::Priority>(getLogger().getLevel()));
	}

	void setLevel(LogLevel lvl) override
	{
		BaseLogger::setLevel(lvl);
		getLogger().setLevel(toPocoLogLevel(lvl));
	}

	void verbose(const std::string& msg) override
	{
		getLogger().trace(msg);
	}

	void debug(const std::string& msg) override
	{
		getLogger().debug(msg);
	}

	void info(const std::string& msg) override
	{
		getLogger().information(msg);
	}

	void warn(const std::string& msg) override
	{
		getLogger().warning(msg);
	}

	void error(const std::string& msg) override
	{
		getLogger().error(msg);
	}

	void critical(const std::string& msg) override
	{
		getLogger().critical(msg);
	}

	void addLoggingChannel(Poco::Channel* newChannel) override
	{
		Poco::Channel* currentChannel = getLogger().getChannel();
		if (currentChannel)
		{
			// If a channel already exists, create a new splitter channel and
			// add the existing and the new channel to the splitter channel.
			// After that set the splitter channel as top channel of the logger.
			Poco::AutoPtr<Poco::SplitterChannel> splitterChannel(new Poco::SplitterChannel, true);
			splitterChannel->addChannel(currentChannel);
			splitterChannel->addChannel(newChannel);
			getLogger().setChannel(splitterChannel);
		}
		else
		{
			// The current channel of the logger is null so just
			// set the new channel as top channel.
			getLogger().setChannel(newChannel);
		}
	}

	void removeLoggingChannel(Poco::Channel* channel) override
	{
		Poco::Channel* currentChannel = getLogger().getChannel();
		if (isSplitterChannel(currentChannel))
		{
			recurse(currentChannel, channel);  // Current channel is splitter
		}
		else if (channel == currentChannel)
		{
			getLogger().setChannel(nullptr);
		}
		else
		{
			/* throw std::runtime_error("Channel can't be released because the Logger does not hold it"); */
			// Do nothing ...
		}
	}

	Poco::Logger& getLogger()
	{
		return Poco::Logger::get(name_);
	}

private:
	bool areSame(Poco::Channel* c1, Poco::Channel* c2)
	{
		return c1 == c2;
	}

	bool isSplitterChannel(Poco::Channel* channel)
	{
		return typeid(channel).name() == typeid(Poco::SplitterChannel).name();
	}

	void recurse(Poco::Channel* currentChannel, Poco::Channel* channel)
	{
		auto* splitter = static_cast<Poco::SplitterChannel*>(currentChannel);
		for (int i = 0; i < splitter->count(); i++)
		{
			auto* subChannelOfSplitter = &splitter[i];
			if (areSame(subChannelOfSplitter, channel))
			{
				std::cout << "remove channel" << LEND;
				splitter->removeChannel(channel);
				return;
			}

			if (isSplitterChannel(subChannelOfSplitter))
			{
				recurse(subChannelOfSplitter, channel);
			}
			else
			{
				// channel is not the channel to remove and
				// also not a SplitterChannel. So do nothing.
				break;
			}
		}
	}
};


namespace deprecated
{
/*!
    Encapsulates the Poco logger
    To ensure the Functionality of old libraries that use the Logger from Poco
    this Class maps to the macro WZLOGGER, if USE_POCO_LEGACY_LOGGER is defined.

    Derives from Poco::Logger

    This class is only defined if the macro USE_POCO_LEGACY_LOGGER is set to 1
*/
class PocoLegacyLogger : public Logger
{
public:
	/*!
	    Creates a logger that internally derives from Poco::Logger

	    @param name     Displayed name of the logger
	    @param lvl      Log level
	    @param chnl     Pointer to a Poco::Channel
	 */
	PocoLegacyLogger(const char* name, LogLevel lvl) : Logger(name, lvl){};

	Poco::LogStream&& write(LogLevel lvl) const
	{
		Poco::LogStream ls(name_);
		switch (lvl)
		{
			case LogLevel::verbose:
				return std::move(ls.trace());
			case LogLevel::debug:
				return std::move(ls.debug());
			case LogLevel::info:
				return std::move(ls.information());
			case LogLevel::warn:
				return std::move(ls.warning());
			case LogLevel::error:
				return std::move(ls.error());
			case LogLevel::critical:
				return std::move(ls.critical());
			case LogLevel::fatal:
				return std::move(ls.fatal());
			case LogLevel::notice:
				return std::move(ls.notice());
			default:
				return std::move(ls);
		}
	}
};
}  // namespace deprecated
}  // namespace wzcpp
#endif  // !__WZCPP_LOGGING_POCOLEGACYLOGGER_H__

#endif  // USE_POCO_LEGACY_LOGGER
