//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

/*
 * This detail file is included in every file of the folder logging.
 * It provides the necessary macros for configuring the loggers.
 */

#ifndef __WZCPP_LOGGING_DETAIL_H__
#define __WZCPP_LOGGING_DETAIL_H__

#if USE_POCO_LEGACY_LOGGER == 1
#include <Poco/Logger.h>
#endif

namespace wzcpp
{
	
/******************************************************************************
 ** Log levels
 *****************************************************************************/
using LOGLEVEL = unsigned int;

enum class LogLevel : LOGLEVEL
{
	verbose  = 0,
	debug    = 1,
	info     = 2,
	notice   = 3,
	warn     = 4,
	error    = 5,
	critical = 6,
	fatal    = 7,
	off      = 8
};

}  // namespace wzcpp

#define LVL_VERBOSE  wzcpp::LogLevel::verbose
#define LVL_DEBUG    wzcpp::LogLevel::debug
#define LVL_INFO     wzcpp::LogLevel::info
#define LVL_WARN     wzcpp::LogLevel::warn
#define LVL_ERROR    wzcpp::LogLevel::error
#define LVL_CRITICAL wzcpp::LogLevel::critical

#if USE_POCO_LEGACY_LOGGER == 1
#define LVL_FATAL  wzcpp::LogLevel::fatal
#define LVL_NOTICE wzcpp::LogLevel::notice
#else
#define LVL_FATAL  wzcpp::LogLevel::critical
#define LVL_NOTICE wzcpp::LogLevel::info
#endif

#define LVL_OFF wzcpp::LogLevel::off


/*!
    The standard format for logging messages

    %v  The actual text to log
    %t  Thread id
    %n  Logger's name
    %l  The log level of the message
    %a  Abbreviated weekday name
    %A  Full weekday name
    %b  Abbreviated month name
    %B  Full month name
    %c  Date and time representation
    %C  Year in 2 digits
    %Y  Year in 4 digits
    %D  Short MM/DD/YY date
    %m  Month 01-12
    %d  Day of month 01-31
    %H  Hours in 24 format  00-23
    %I  Hours in 12 format  01-12
    %M  Minutes 00-59
    %S  Seconds 00-59
    %e  Millisecond part of the current second 000-999
    %f  Microsecond part of the current second 000000-999999
    %F  Nanosecond part of the current second 000000000-999999999
    %p  AM/PM|"AM"|
    %T  or `%X`|ISO 8601 time format (HH:MM:SS), equivalent to %H:%M:%S
    %^  start color range (can be used only once)
    %$  end color range (for example %^[+++]%$ %v)
*/
#ifndef WZLOGPATTERN
#define WZLOGPATTERN "%Y-%m-%d %T.%e %^%-7l%$ [%n] %v"
#endif
/******************************************************************************
 ** Logger creator
 *****************************************************************************/
#define WZ_DEF_LOGNAME(Name)                       \
	namespace                                      \
	{                                              \
	extern const char wzfClassLoggerName[] = Name; \
	}

// If USE_POCO_LEGACY_LOGGER is defined the function used
// for creating a PocoLegacyLogger  is mapped to the WZLOG macro

#if defined USE_POCO_LEGACY_LOGGER && USE_POCO_LEGACY_LOGGER == 1
#define WZLOGGER(name, lvl)          \
	wzcpp::PocoLegacyLogger wzlogger \
	{                                \
		name, lvl                    \
	}
#define WZLOGGERDEFAULT() auto& wzlogger = wzcpp::PocoLegacyLogger::root()
#define WZ_DEF_LOGGER()   wzcpp::PocoLegacyLogger(::wzfClassLoggerName, LVL_VERBOSE)
#else
// clang-format off
#define WZLOGGER(name, lvl) wzcpp::WzLogger wzlogger{name, lvl}
#define WZLOGGERDEFAULT()     wzcpp::WzLogger wzlogger{"default", LVL_DEBUG}
#define WZ_DEF_LOGGER() wzcpp::WzLogger(::wzfClassLoggerName, LVL_VERBOSE)
// clang-format on
#endif

/******************************************************************************
 ** Logging macros
 *****************************************************************************/

//---------------------------------------------------------
// Poco legacy
//
// These legacy macros map to their counterparts in KDrive
// See kdrive/include/kdrive/LoggingEnabled.h
//---------------------------------------------------------
#if defined USE_POCO_LEGACY_LOGGER && USE_POCO_LEGACY_LOGGER == 1
// clang-format off

//
// Verbose
//
#define wzlog_v(logger, msg)                                poco_trace(logger.getLogger(), msg)                           
#define wzlog_v_f1(logger, fmt, arg1)                       poco_trace_f1(logger.getLogger(), fmt, arg1)                  
#define wzlog_v_f2(logger, fmt, arg1, arg2)                 poco_trace_f2(logger.getLogger(), fmt, arg1, arg2)            
#define wzlog_v_f3(logger, fmt, arg1, arg2, arg3)           poco_trace_f3(logger.getLogger(), fmt, arg1, arg2, arg3)      
#define wzlog_v_f4(logger, fmt, arg1, arg2, arg3, arg4)     poco_trace_f4(logger.getLogger(), fmt, arg1, arg2, arg3, arg4)
//
// Fatal
//
#define wzlog_f(logger, msg)                                poco_fatal(logger.getLogger(), msg)
#define wzlog_f_f1(logger, fmt, arg1)                       poco_fatal_f1(logger.getLogger(), fmt, arg1)
#define wzlog_f_f2(logger, fmt, arg1, arg2)                 poco_fatal_f2(logger.getLogger(), fmt, arg1, arg2)
#define wzlog_f_f3(logger, fmt, arg1, arg2, arg3)           poco_fatal_f3(logger.getLogger(), fmt, arg1, arg2, arg3)
#define wzlog_f_f4(logger, fmt, arg1, arg2, arg3, arg4)     poco_fatal_f4(logger.getLogger(), fmt, arg1, arg2, arg3, arg4)
//
// Critical
//
#define wzlog_c(logger, msg)                                poco_critical(logger.getLogger(), msg)
#define wzlog_c_f1(logger, fmt, arg1)                       poco_critical_f1(logger.getLogger(), fmt, arg1)
#define wzlog_c_f2(logger, fmt, arg1, arg2)                 poco_critical_f2(logger.getLogger(), fmt, arg1, arg2)
#define wzlog_c_f3(logger, fmt, arg1, arg2, arg3)           poco_critical_f3(logger.getLogger(), fmt, arg1, arg2, arg3)
#define wzlog_c_f4(logger, fmt, arg1, arg2, arg3, arg4)     poco_critical_f4(logger.getLogger(), fmt, arg1, arg2, arg3, arg4)
//
// Error
//
#define wzlog_e(logger, msg)                                poco_error(logger.getLogger(), msg)
#define wzlog_e_f1(logger, fmt, arg1)                       poco_error_f1(logger.getLogger(), fmt, arg1)
#define wzlog_e_f2(logger, fmt, arg1, arg2)                 poco_error_f2(logger.getLogger(), fmt, arg1, arg2)
#define wzlog_e_f3(logger, fmt, arg1, arg2, arg3)           poco_error_f3(logger.getLogger(), fmt, arg1, arg2, arg3)
#define wzlog_e_f4(logger, fmt, arg1, arg2, arg3, arg4)     poco_error_f4(logger.getLogger(), fmt, arg1, arg2, arg3, arg4)
//
// Warning
//
#define wzlog_w(logger, msg)                                poco_warning(logger.getLogger(), msg)
#define wzlog_w_f1(logger, fmt, arg1)                       poco_warning_f1(logger.getLogger(), fmt, arg1)
#define wzlog_w_f2(logger, fmt, arg1, arg2)                 poco_warning_f2(logger.getLogger(), fmt, arg1, arg2)
#define wzlog_w_f3(logger, fmt, arg1, arg2, arg3)           poco_warning_f3(logger.getLogger(), fmt, arg1, arg2, arg3)
#define wzlog_w_f4(logger, fmt, arg1, arg2, arg3, arg4)     poco_warning_f4(logger.getLogger(), fmt, arg1, arg2, arg3, arg4)
//
// Notice
//
#define wzlog_n(logger, msg)                                poco_notice(logger.getLogger(), msg)
#define wzlog_n_f1(logger, fmt, arg1)                       poco_notice_f1(logger.getLogger(), fmt, arg1)
#define wzlog_n_f2(logger, fmt, arg1, arg2)                 poco_notice_f2(logger.getLogger(), fmt, arg1, arg2)
#define wzlog_n_f3(logger, fmt, arg1, arg2, arg3)           poco_notice_f3(logger.getLogger(), fmt, arg1, arg2, arg3)
#define wzlog_n_f4(logger, fmt, arg1, arg2, arg3, arg4)     poco_notice_f4(logger.getLogger(), fmt, arg1, arg2, arg3, arg4)
//
// Information
//
#define wzlog_i(logger, msg)                                poco_information(logger.getLogger(), msg)
#define wzlog_i_f1(logger, fmt, arg1)                       poco_information_f1(logger.getLogger(), fmt, arg1)
#define wzlog_i_f2(logger, fmt, arg1, arg2)                 poco_information_f2(logger.getLogger(), fmt, arg1, arg2)
#define wzlog_i_f3(logger, fmt, arg1, arg2, arg3)           poco_information_f3(logger.getLogger(), fmt, arg1, arg2, arg3)
#define wzlog_i_f4(logger, fmt, arg1, arg2, arg3, arg4)     poco_information_f4(logger.getLogger(), fmt, arg1, arg2, arg3, arg4)
//
// Debug
//
#define wzlog_d(logger, msg)                                poco_debug(logger.getLogger(), msg)
#define wzlog_d_f1(logger, fmt, arg1)                       poco_debug_f1(logger.getLogger(), fmt, arg1)
#define wzlog_d_f2(logger, fmt, arg1, arg2)                 poco_debug_f2(logger.getLogger(), fmt, arg1, arg2)
#define wzlog_d_f3(logger, fmt, arg1, arg2, arg3)           poco_debug_f3(logger.getLogger(), fmt, arg1, arg2, arg3)
#define wzlog_d_f4(logger, fmt, arg1, arg2, arg3, arg4)     poco_debug_f4(logger.getLogger(), fmt, arg1, arg2, arg3, arg4)
//
// Trace
//
#define wzlog_t(logger, msg)                                poco_trace(logger.getLogger(), msg)
#define wzlog_t_f1(logger, fmt, arg1)                       poco_trace_f1(logger.getLogger(), fmt, arg1)
#define wzlog_t_f2(logger, fmt, arg1, arg2)                 poco_trace_f2(logger.getLogger(), fmt, arg1, arg2)
#define wzlog_t_f3(logger, fmt, arg1, arg2, arg3)           poco_trace_f3(logger.getLogger(), fmt, arg1, arg2, arg3)
#define wzlog_t_f4(logger, fmt, arg1, arg2, arg3, arg4)     poco_trace_f4(logger.getLogger(), fmt, arg1, arg2, arg3, arg4)
#endif
// Stream logger
#define wzlogs_v(msg) {std::ostringstream s; s << msg; wzlogger.stream(LVL_VERBOSE, s);}
#define wzlogs_d(msg) {std::ostringstream s; s << msg; wzlogger.stream(LVL_DEBUG, s);}
#define wzlogs_n(msg) {std::ostringstream s; s << msg; wzlogger.stream(LVL_NOTICE, s);}
#define wzlogs_i(msg) {std::ostringstream s; s << msg; wzlogger.stream(LVL_INFO, s);}
#define wzlogs_w(msg) {std::ostringstream s; s << msg; wzlogger.stream(LVL_WARN, s);}
#define wzlogs_e(msg) {std::ostringstream s; s << msg; wzlogger.stream(LVL_ERROR, s);}
#define wzlogs_c(msg) {std::ostringstream s; s << msg; wzlogger.stream(LVL_CRITICAL, s);}
#define wzlogs_f(msg) {std::ostringstream s; s << msg; wzlogger.stream(LVL_FATAL, s);}

//---------------------------------------------------------
// Wz logging
//---------------------------------------------------------
#if !defined USE_POCO_LEGACY_LOGGER || USE_POCO_LEGACY_LOGGER != 1
// Function logger
#define wzlog_v(logger, msg)                                logger.verbose(msg)
#define wzlog_v_f1(logger, fmt, arg1)                       logger.verbose(wzlf(fmt, arg1))
#define wzlog_v_f2(logger, fmt, arg1, arg2)                 logger.verbose(wzlf(fmt, arg1, arg2))
#define wzlog_v_f3(logger, fmt, arg1, arg2, arg3)           logger.verbose(wzlf(fmt, arg1, arg2, arg3))
#define wzlog_v_f4(logger, fmt, arg1, arg2, arg3, arg4)     logger.verbose(wzlf(fmt, arg1, arg2, arg3, arg4))

#define wzlog_d(logger, msg)                                logger.debug(msg)
#define wzlog_d_f1(logger, fmt, arg1)                       logger.debug(wzlf(fmt, arg1))
#define wzlog_d_f2(logger, fmt, arg1, arg2)                 logger.debug(wzlf(fmt, arg1, arg2))
#define wzlog_d_f3(logger, fmt, arg1, arg2, arg3)           logger.debug(wzlf(fmt, arg1, arg2, arg3))
#define wzlog_d_f4(logger, fmt, arg1, arg2, arg3, arg4)     logger.debug(wzlf(fmt, arg1, arg2, arg3, arg4))

#define wzlog_n(logger, msg)                                logger.info(msg)
#define wzlog_n_f1(logger, fmt, arg1)                       logger.info(wzlf(fmt, arg1))
#define wzlog_n_f2(logger, fmt, arg1, arg2)                 logger.info(wzlf(fmt, arg1, arg2))
#define wzlog_n_f3(logger, fmt, arg1, arg2, arg3)           logger.info(wzlf(fmt, arg1, arg2, arg3))
#define wzlog_n_f4(logger, fmt, arg1, arg2, arg3, arg4)     logger.info(wzlf(fmt, arg1, arg2, arg3, arg4))

#define wzlog_i(logger, msg)                                logger.info(msg)
#define wzlog_i_f1(logger, fmt, arg1)                       logger.info(wzlf(fmt, arg1))
#define wzlog_i_f2(logger, fmt, arg1, arg2)                 logger.info(wzlf(fmt, arg1, arg2))
#define wzlog_i_f3(logger, fmt, arg1, arg2, arg3)           logger.info(wzlf(fmt, arg1, arg2, arg3))
#define wzlog_i_f4(logger, fmt, arg1, arg2, arg3, arg4)     logger.info(wzlf(fmt, arg1, arg2, arg3, arg4))

#define wzlog_w(logger, msg)                                logger.warn(msg)
#define wzlog_w_f1(logger, fmt, arg1)                       logger.warn(wzlf(fmt, arg1))
#define wzlog_w_f2(logger, fmt, arg1, arg2)                 logger.warn(wzlf(fmt, arg1, arg2))
#define wzlog_w_f3(logger, fmt, arg1, arg2, arg3)           logger.warn(wzlf(fmt, arg1, arg2, arg3))
#define wzlog_w_f4(logger, fmt, arg1, arg2, arg3, arg4)     logger.warn(wzlf(fmt, arg1, arg2, arg3, arg4))

#define wzlog_e(logger, msg)                                logger.error(msg)
#define wzlog_e_f1(logger, fmt, arg1)                       logger.error(wzlf(fmt, arg1))
#define wzlog_e_f2(logger, fmt, arg1, arg2)                 logger.error(wzlf(fmt, arg1, arg2))
#define wzlog_e_f3(logger, fmt, arg1, arg2, arg3)           logger.error(wzlf(fmt, arg1, arg2, arg3))
#define wzlog_e_f4(logger, fmt, arg1, arg2, arg3, arg4)     logger.error(wzlf(fmt, arg1, arg2, arg3, arg4))

#define wzlog_c(logger, msg)                                logger.critical(msg)
#define wzlog_c_f1(logger, fmt, arg1)                       logger.critical(wzlf(fmt, arg1))
#define wzlog_c_f2(logger, fmt, arg1, arg2)                 logger.critical(wzlf(fmt, arg1, arg2))
#define wzlog_c_f3(logger, fmt, arg1, arg2, arg3)           logger.critical(wzlf(fmt, arg1, arg2, arg3))
#define wzlog_c_f4(logger, fmt, arg1, arg2, arg3, arg4)     logger.critical(wzlf(fmt, arg1, arg2, arg3, arg4))

#define wzlog_f(logger, msg)                                logger.critical(msg)
#define wzlog_f_f1(logger, fmt, arg1)                       logger.critical(wzlf(fmt, arg1))
#define wzlog_f_f2(logger, fmt, arg1, arg2)                 logger.critical(wzlf(fmt, arg1, arg2))
#define wzlog_f_f3(logger, fmt, arg1, arg2, arg3)           logger.critical(wzlf(fmt, arg1, arg2, arg3))
#define wzlog_f_f4(logger, fmt, arg1, arg2, arg3, arg4)     logger.critical(wzlf(fmt, arg1, arg2, arg3, arg4))
#endif

// clang-format on
#define LEND "\n"
#endif  // !__WZCPP_LOGGING_DETAIL_H__
