//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#ifndef __WZCPP_UTILS_BYTEARRAY_H__
#define __WZCPP_UTILS_BYTEARRAY_H__

#include <array>
#include <iomanip>
#include <iostream>
#include <regex>
#include <sstream>
#include <stdexcept>
#include <string>
#include <vector>

namespace wzcpp
{
template<int length>
class ByteArray
{
public:
	using DataBuffer = std::array<unsigned char, length>;

	/*!
	    Creates an array with a fixes size of bytes assigned with zeros.
	*/
	ByteArray() : dataBuffer_()
	{
		for (size_t i = 0; i < length; i++)
		{
			dataBuffer_[i] = 0;
		}
	}

	/*!
	    CopyConstructor
	*/
	ByteArray(ByteArray& other) : dataBuffer_(other.dataBuffer_)
	{
	}


	/*!
	    MoveConstructor
	*/
	ByteArray(ByteArray&& other) : dataBuffer_(std::move(other.dataBuffer_))
	{
	}


	/*!
	    Creates an array with 6 bytes assign with zeros.
	    @param data An unsigned char array of 6 bytes length
	*/
	explicit ByteArray(const DataBuffer& data) : dataBuffer_(data)
	{
	}

	/*!
	    Creates from a vector.

	    @param data is a vector with 6 bytes
	    @exception  std::range_error buffer is null or bufferLength is not
	                SerialNumberLength (6 bytes).
	*/
	explicit ByteArray(const std::vector<unsigned char>& data)
	{
		if (data.size() != length)
		{
			// TODO(mschaetz): Create RangeExceptionClass
			// throw RangeException("Expected 6 bytes");
			std::ostringstream stream;
			stream << "Expected " << length << " bytes";
			throw std::range_error(stream.str());
		}

		std::copy(data.begin(), data.end(), dataBuffer_.begin());
	}

	/*!
	    Creates from a buffer.

	    @param serialNumber is an array with 6 bytes
	    @param bufferLength is the array length of the buffer (must be 6)
	    @exception          std::range_error buffer is null or bufferLength is not
	                        SerialNumberLength (6 bytes).
	*/
	ByteArray(const unsigned char* buffer, std::size_t bufferLength)
	{
		if (!buffer || bufferLength < length)
		{
			// throw RangeException("Expected 6 bytes");
			std::ostringstream stream;
			stream << "Expected " << length << " bytes";
			throw std::range_error(stream.str());
		}

		std::memcpy(&dataBuffer_.at(0), buffer, bufferLength);
	}

	/*!
	    Creates from a string.
	*/
	explicit ByteArray(const std::string& hexString) : ByteArray()
	{
		// Detects any alphanumerical string with hex identifiers
		// with two characters inside strings like
		// 0102fdFE
		// 0x0102fDFe
		// 0x01 0x02 0xfD 0xFe
		auto regExpression = std::regex("([\\dA-Fa-f]{2})");

		auto searchFoundMatches = std::regex_search(hexString, regExpression);

		if (!searchFoundMatches)
		{
			throw std::runtime_error("hexString does not match the regular expression '([\\dA-Fa-f]{2})'");
		}

		auto begin = std::sregex_iterator(hexString.begin(), hexString.end(), regExpression);
		auto end   = std::sregex_iterator();
		std::ostringstream ss;

		std::size_t bufferIndex = 0;
		for (std::sregex_iterator iter = begin; iter != end; iter++)
		{
			std::smatch match = *iter;
			ss << std::hex << match.str();

			if (bufferIndex >= dataBuffer_.size())
			{
				// Index indicates an element that exceeds the size of the data buffer.
				std::ostringstream errStream("Databuffer can hold only ");
				errStream << length << " elements, but a " << length + 1 << "th element was found.";
				throw std::range_error(errStream.str());
			}

			dataBuffer_[bufferIndex++] = static_cast<unsigned char>(std::stoi(ss.str(), nullptr, 16));

			ss.seekp(0);
		}

		if ((bufferIndex) < dataBuffer_.size())
		{
			// There are less matches than 4. So that means less than 4 byte where
			// written in the data buffer.
			std::ostringstream stream("Databuffer contains less than ");
			stream << length << " byte";
			throw std::range_error(stream.str());
		}
	}

	/*!
	    Destroy the serial number.
	*/
	virtual ~ByteArray() = default;

	/*!
	    Assignment operator
	*/
	ByteArray& operator=(const ByteArray& other)
	{
		if (this != &other)
		{
			dataBuffer_ = other.dataBuffer_;
		}
		return *this;
	}

	/*!
	    Move assignment operator
	*/
	ByteArray& operator=(ByteArray&& other)
	{
		if (this != &other)
		{
			dataBuffer_ = std::move(other.dataBuffer_);
		}
		return *this;
	}

	/*!
	    Equality operator
	*/
	bool operator==(const ByteArray& other) const
	{
		return (dataBuffer_ == other.dataBuffer_);
	}

	/*!
	    Inequality operator
	*/
	bool operator!=(const ByteArray& other) const
	{
		return (dataBuffer_ != other.dataBuffer_);
	}

	/*!
	    Less than operator
	*/
	bool operator<(const ByteArray& other) const
	{
		return (dataBuffer_ < other.dataBuffer_);
	}

	/*!
	    Less than or equal operator
	*/
	bool operator<=(const ByteArray& other) const
	{
		return (dataBuffer_ <= other.dataBuffer_);
	}

	/*!
	    Greater than operator
	*/
	bool operator>(const ByteArray& other) const
	{
		return (dataBuffer_ > other.dataBuffer_);
	}

	/*!
	    Greater than or equal operator
	*/
	bool operator>=(const ByteArray& other) const
	{
		return (dataBuffer_ >= other.dataBuffer_);
	}

	/*!
	    Pre-increment operator
	    \note If the value is already FF FF FF FF FF FF, than the value is after the
	   increment 00 00 00 00 00 00!
	*/
	virtual ByteArray& operator++()
	{
		auto iter = dataBuffer_.rbegin();
		for (; iter != dataBuffer_.rend(); ++iter)
		{
			unsigned char& value = *iter;
			if (value == 0xFF)
			{
				value = 0;
			}
			else
			{
				++value;
				break;
			}
		}

		return *this;
	}

	/*!
	    Pre-decrement operator
	    \note If the value is 00 00 00 00 00 00, than the value is after the decrement  FF
	   FF FF FF FF FF!
	*/
	ByteArray& operator--()
	{
		auto iter = dataBuffer_.rbegin();
		for (; iter != dataBuffer_.rend(); ++iter)
		{
			unsigned char& value = *iter;
			if (value == 0x00)
			{
				value = 0xFF;
			}
			else
			{
				--value;
				break;
			}
		}

		return *this;
	}

	/*!
	    Plus assignment operator
	*/
	ByteArray& operator+=(unsigned int increment)
	{
		for (unsigned int i = 0; i < increment; ++i)
		{
			operator++();
		}
		return *this;
	}

	/*!
	    Minus assignment operator
	*/
	ByteArray& operator-=(unsigned int decrement)
	{
		for (unsigned int i = 0; i < decrement; ++i)
		{
			operator--();
		}
		return *this;
	}


	/*!
	    Conversion operator for vector
	*/
	operator std::vector<unsigned char>() const
	{
		return toVector();
	}

	/*!
	    Returns a reference to the internal byte array.
	*/
	DataBuffer& getDataBuffer()
	{
		return dataBuffer_;
	}

	/*!
	    Returns a const reference to the internal byte array.
	*/
	const DataBuffer& getDataBuffer() const
	{
		return dataBuffer_;
	}

	/*!
	    Return a string with the formatted buffer.
	    @param placeholder is the placeholder between the hex values. The default is " ".
	*/
	std::string toString(const std::string& placeholder = " ") const
	{
		std::ostringstream stream;

		for (unsigned int index = 0; index < dataBuffer_.size(); ++index)
		{
			if (index)
			{
				stream << placeholder;
			}
			stream << std::uppercase << std::setfill('0') << std::setw(2) << std::hex
			       << static_cast<uint16_t>(dataBuffer_.at(index));
		}

		return stream.str();
	}

	std::string toZeroXString() const
	{
		std::string str = "0x";
		return str + toString(" 0x");
	}

	/*!
	    Return a vector with the serial number.
	*/
	std::vector<unsigned char> toVector() const
	{
		std::vector<unsigned char> data(dataBuffer_.size());
		std::copy(dataBuffer_.begin(), dataBuffer_.end(), data.begin());

		return data;
	}

protected:
	DataBuffer dataBuffer_;
};  // namespace utils

/**********************************************************************************
 * FourByteArray
 **********************************************************************************/

class FourByteArray : public ByteArray<4>
{
public:
	using ByteArray::ByteArray;

	/*!
	    Creates a array with 4 bytes assign with zeros.
	*/
	FourByteArray() : ByteArray<4>()
	{
	}

	/*!
	    Copy constructor
	*/
	FourByteArray(const FourByteArray& byteArray)
	:ByteArray<4>()
	{
		dataBuffer_ = byteArray.getDataBuffer();
	}

	/*!
	    Move constructor
	*/
	FourByteArray(FourByteArray&& other)
	{
		dataBuffer_ = std::move(other.dataBuffer_);
	}

	FourByteArray(const DataBuffer& data) : ByteArray<4>(data)
	{
	}

	/*!
	    Creates from a vector.

	    \param[in] data is a vector with 4 bytes
	*/
	explicit FourByteArray(const std::vector<unsigned char>& data) : ByteArray<4>(data)
	{
	}

	/*!
	    Creates from a buffer.

	    \param[in] serialNumber is a array with 4 bytes
	    \param[in] bufferLength is the array length of the buffer (must be 4)
	    \exception Poco::RangeException buffer is null or bufferLength is not Length (4
	   bytes).
	*/
	FourByteArray(const unsigned char* buffer, std::size_t bufferLength) : ByteArray(buffer, bufferLength)
	{
	}

	explicit FourByteArray(const std::string& hexString) : ByteArray(hexString)
	{
	}

	/*!
	    Creates with specified values.
	    \param[in] byte0 Byte 0 of the data
	    \param[in] byte1 Byte 1 of the data
	    \param[in] byte2 Byte 2 of the data
	    \param[in] byte3 Byte 3 of the data
	*/
	FourByteArray(unsigned char byte0, unsigned char byte1, unsigned char byte2, unsigned char byte3)
	{
		dataBuffer_.at(0) = byte0;
		dataBuffer_.at(1) = byte1;
		dataBuffer_.at(2) = byte2;
		dataBuffer_.at(3) = byte3;
	}

	/*!
	    Creates from a 4 byte integer.
	 */
	FourByteArray(uint32_t numValue)
	{
		dataBuffer_.at(0) = (numValue >> 24) & 0xFF;
		dataBuffer_.at(1) = (numValue >> 16) & 0xFF;
		dataBuffer_.at(2) = (numValue >> 8) & 0xFF;
		dataBuffer_.at(3) = numValue & 0xFF;
	}

	~FourByteArray() = default;

	FourByteArray& operator=(const FourByteArray& other)
	{
		if (*this != other) this->dataBuffer_ = other.dataBuffer_;
		return *this;
	}

	FourByteArray& operator=(FourByteArray&& other)
	{
		if (*this != other) this->dataBuffer_ = std::move(other.dataBuffer_);
		return *this;
	}

	/*!
	    Minus operator
	*/
	uint32_t operator-(const FourByteArray& other) const
	{
		uint32_t span      = 0;
		uint32_t aryint1   = 0;
		uint32_t aryint2   = 0;
		unsigned int index = 0;

		// Convert 4 Bytes arrays to 32 bit integer
		for (int i = (4 - 1) * 8; i >= 0; i -= 8)
		{
			aryint1 |= static_cast<uint32_t>(dataBuffer_.at(index)) << i;
			aryint2 |= static_cast<uint32_t>(other.getDataBuffer().at(index)) << i;
			index++;
		}
		span = aryint1 - aryint2;

		return span;
	}
};

/**********************************************************************************
 * SixByteArray
 **********************************************************************************/

class SixByteArray : public ByteArray<6>
{
public:
	/*!
	    Creates an array with 6 bytes assign with zeros.
	*/
	SixByteArray() : ByteArray(){};

	SixByteArray(const SixByteArray& other) : ByteArray(other){};

	SixByteArray(SixByteArray&& other) : ByteArray(other){};

	/*!
	    Creates an array with 6 bytes assign with zeros.
	    @param data An unsigned char array of 6 bytes length
	*/
	SixByteArray(const DataBuffer& data) : ByteArray(data)
	{
	}

	/*!
	    Creates from a vector.

	    @param data is a vector with 6 bytes
	    @exception  std::range_error buffer is null or bufferLength is not
	                SerialNumberLength (6 bytes).
	*/
	SixByteArray(const std::vector<unsigned char>& data) : ByteArray(data)
	{
	}

	/*!
	    Creates from a buffer.

	    @param serialNumber is an array with 6 bytes
	    @param bufferLength is the array length of the buffer (must be 6)
	    @exception          std::range_error buffer is null or bufferLength is not
	                        SerialNumberLength (6 bytes).
	*/
	SixByteArray(const unsigned char* buffer, std::size_t bufferLength) : ByteArray(buffer, bufferLength)
	{
	}

	/*!
	    Creates from a string.
	*/
	SixByteArray(const std::string& hexString) : ByteArray(hexString)
	{
	}

	/*!
	    Creates with specified values.
	    @param byte0 Byte 0 of the data
	    @param byte1 Byte 1 of the data
	    @param byte2 Byte 2 of the data
	    @param byte3 Byte 3 of the data
	    @param byte4 Byte 4 of the data
	    @param byte5 Byte 5 of the data
	*/
	SixByteArray(unsigned char byte0,
	             unsigned char byte1,
	             unsigned char byte2,
	             unsigned char byte3,
	             unsigned char byte4,
	             unsigned char byte5)
	{
		dataBuffer_.at(0) = byte0;
		dataBuffer_.at(1) = byte1;
		dataBuffer_.at(2) = byte2;
		dataBuffer_.at(3) = byte3;
		dataBuffer_.at(4) = byte4;
		dataBuffer_.at(5) = byte5;
	}

	/*!
	    Creates from an 8 byte integer.
	 */
	SixByteArray(uint64_t numValue)
	{
		dataBuffer_.at(0) = (numValue >> 40) & 0xFF;
		dataBuffer_.at(1) = (numValue >> 32) & 0xFF;
		dataBuffer_.at(2) = (numValue >> 24) & 0xFF;
		dataBuffer_.at(3) = (numValue >> 16) & 0xFF;
		dataBuffer_.at(4) = (numValue >> 8) & 0xFF;
		dataBuffer_.at(5) = numValue & 0xFF;
	}

	~SixByteArray() = default;


	SixByteArray& operator=(const SixByteArray& other)
	{
		if (*this != other) this->dataBuffer_ = other.dataBuffer_;
		return *this;
	}

	SixByteArray& operator=(SixByteArray&& other)
	{
		if (*this != other) this->dataBuffer_ = std::move(other.dataBuffer_);
		return *this;
	}

	/*!
	    Minus operator
	*/
	uint64_t operator-(const SixByteArray& other) const
	{
		uint64_t span      = 0;
		uint64_t macint1   = 0;
		uint64_t macint2   = 0;
		unsigned int index = 0;
		// Convert 6 Bytes arrays to 64 bit integer
		for (int i = 40; i >= 0; i -= 8)
		{
			macint1 |= static_cast<uint64_t>(dataBuffer_.at(index)) << i;
			macint2 |= static_cast<uint64_t>(other.getDataBuffer().at(index)) << i;
			index++;
		}
		span = macint1 - macint2;

		if (span == 0xFFFFFFFFFFFFFFFF)
		{
			// Shift 16 bit to the right to create the
			// max value of a six byte array
			span = span >> 0x10;
		}
		return span;
	}
};
}  // namespace wzcpp

#endif  // __WZCPP_UTILS_BYTEARRAY_H__
