//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#ifndef __BAOSLIB_INTERFACE_H__
#define __BAOSLIB_INTERFACE_H__

#include "baos/config.h"

#include "BlockReader.h"
#include "baos/device/BaosDeviceDataTypes.h"
#include "baos/device/BaosDevicesMeta.h"
#include "baos/protocol/Defines.h"

#include "wzcpp/features/algorithm/append.h"
#include "wzcpp/features/logging.h"
#include "wzcpp/features/utils/ByteArray.h"
#include "wzcpp/features/utils/Memory.h"

/// @cond DOCUMENT_STL
#include <memory>
#include <optional>
#include <utility>
/// @endcond

/*!
    @file
    @brief The high level API for BAOS devices
*/

namespace wz::baos
{
// Forward declaration
class BaosConnection;
using BaosConnection_UPtr = std::unique_ptr<BaosConnection>;  //!< Alias for a std::unique pointer of a BaosConnection


/**************************************************
 * BaosDevice
 * ***********************************************/

/*!
    @brief This class represents a BAOS device

    Maintining a BaosConnection to a physical BAOS device all interactions and information with and about said devices
   are stored in this class.

 */
class BAOSLIB_EXPORT BaosDevice
{
public:
	using ServerItemIndicationCb = std::function<void(
	    const protocol::ServerItems&)>;  //!< The function signature for the callback on changed server items
	using DatapointValueCb       = std::function<void(
        const protocol::DatapointValues&)>;  //!< The function signature for the callback on changed datapoint values

	BaosDevice()          = default;
	virtual ~BaosDevice() = default;

	virtual void close(); /*!< Close the connection to the baos device */

	/************************ Information about the device ************************/
	/*!
	    @name Device information
	    Information about the device via serveritems
	 */
	///@{
	/*!
	    @brief Get the information about a connected BAOS device that is never changed. Like serial number

	    @param reload true if the information should be directly loaded from the device, false(default) for using the
	   cache if possible
	    @return A structure containing the static information
	 */
	DeviceStaticInfo getStaticInfo(bool reload = false);


	/*!
	    @brief Get the information about a connected BAOS device that is typically only changed during configuration.

	    @param reload true if the information should be directly loaded from the device, false(default) for using the
	   cache if possible
	    @return A structure containing the config information
	*/
	DeviceConfigInfo getConfigInfo(bool reload = false);

	/*!
	    @brief Get the information about a connected BAOS device that may change during runtime.

	    @param reload true if the information should be directly loaded from the device, false(default) for using the
	   cache if possible
	    @return A structure containing the runtime information
	*/
	DeviceRuntimeInfo getRuntimeInfo(bool reload = false);

	/*!
	    @brief Set the a bool value to a server item

	    @param id The id of the serveritem
	    @param value The new value to set
	 */
	void setBoolServerItem(protocol::ServerItemId id, bool value);

	///@}

	/************************ Parameter bytes ***************************************/
	/*!
	    @name Parameter bytes related methods
	    Methods to retrieve the parameter bytes from the device
	 */
	///@{

	/*!
	    @brief Get the parameter bytes
	    If caching is enabled, the result of this call will be stored in cache
	    @param byteIndex The first bytes to retrieve(1 based index)
	    @param numberOfBytes The amount of bytes to retrieve
	    @param reload [opt] If true bypass the cache and force a reload from the device, defaults to false
	    @return ParameterBytes
	 */
	protocol::ParameterBytes getParameterBytes(std::uint16_t byteIndex     = 1,
	                                           std::uint16_t numberOfBytes = UINT16_MAX,
	                                           bool reload                 = false);

	/*!
	    @brief Load all the parameter bytes and cache them.

	    To detemine how many bytes are available the serverItem maxParameterBytes is used. If this is not available load
	    1-65535

	    @param reload [opt] If true bypass the cache and force a reload from the device, defaults to false
	    @return The requested parameter bytes
	 */
	protocol::ParameterBytes getAllParameterBytes(bool reload = false);
	///@}

	/**************************** Datapoints *****************************************/
	/*!
	    @name Datapoint related methods
	    All methods for setting and retrieving information and values of datapoints
	 */
	///@{
	/*~~~~~~~~~~~~~~~~~~~~~~~~~~~ Datapoint configurations ~~~~~~~~~~~~~~~~~~~~~~~~~~*/

	/*!
	    @brief Get the datapoint configuration of a single datapoint
	    If caching is enabled, the result of this call will be stored in cache
	    @param dpId The 1 based id of the datapoint to get
	    @param withDescStrings If the description string should be loaded too or not
	    @param reload [opt] If true bypass the cache and force a reload from the device, defaults to false
	    @return The DatapointConfiguration for the given id or an empty optional of not value could be retrieved.
	*/
	std::optional<DatapointConfig> getDpConfiguration(std::uint16_t dpId   = 1,
	                                                  bool withDescStrings = true,
	                                                  bool reload          = false);

	/*!
	    @brief Get the datapoint configurations
	    If caching is enabled, the result of this call will be stored in cache
	    @param startDpId The 1 based id of the first datapoint to get
	    @param endDpId The 1 based id of the last datapoint to get
	    @param withDescStrings If the description string should be loaded too or not
	    @param reload [opt] If true bypass the cache and force a reload from the device, defaults to false
	    @return DatapointConfigurations
	*/
	DatapointConfigurations getDpConfigurations(std::uint16_t startDpId = 1,
	                                            std::uint16_t endDpId   = UINT16_MAX,
	                                            bool withDescStrings    = true,
	                                            bool reload             = false);


	/*!
	    @brief Load all the datapoint configurations and cache them.

	    To detemine how many datapoints are available the serverItem maxDatapoints is used. If this is not available
	   load 1-65535.
	    @param reload [opt] If true bypass the cache and force a reload from the device, defaults to false
	    @return DatapointConfigurations
	 */
	DatapointConfigurations getAllDpConfigurations(bool reload = false);
	/*~~~~~~~~~~~~~~~~~~~~~~~~~~~ Datapoint values ~~~~~~~~~~~~~~~~~~~~~~~~~~*/

	/*!
	    @brief Set the value of a single datapoint

	    @param dpId The 1 based id of the datapoint
	    @param dpData The new value to set
	    @param command [in,opt] What type of change is requested
	 */
	void setDpValue(std::uint16_t dpId,
	                Buffer dpData,
	                protocol::DatapointSetCommand command = protocol::DatapointSetCommand::SetNewValueAndSendOnBus);

	/*!
	    @brief Set the value of multiple datapoints

	    @param values a map with the datapoint ids as keys and the values as entries
	    @param command [in,opt] What type of change is requested
	 */
	void setDpValues(std::map<std::uint16_t, Buffer>&& values,
	                 protocol::DatapointSetCommand command = protocol::DatapointSetCommand::SetNewValueAndSendOnBus);

	/*!
	    @brief Triggers a read request for the datapoint on the bus

	    @param dpId The 1 based id of the datapoint
	 */
	void readDpValueFromBus(std::uint16_t dpId);

	/*!
	    @brief Triggers read requests for a set of datapoints on the bus

	    @param dpIds A vector containing the 1 based datapoints ids, for which the read request should be send out
	 */
	void readDpValuesFromBus(std::vector<std::uint16_t>&& dpIds);


	/*!
	    @brief Get the value of a single datapoint as raw bytes

	    @param dpId The 1 based id of the datapoint
	    @param reload If true bypass the cache and force a reload from the device
	    @return DatapointValue
	 */
	std::optional<protocol::DatapointValue> getDpValueRaw(std::uint16_t dpId, bool reload = false);

	/*!
	    @brief Get the value of a single datapoint as type T

	    @tparam T The type to which the raw Bytes should be converted to
	    @param dpId The 1 based id of the datapoint
	    @param reload If true bypass the cache and force a reload from the device
	 */
	template<class T>
	std::optional<T> getDatapointValue(std::uint16_t dpId, bool reload = false)
	{
		auto value = getDpValueRaw(dpId);
		if (value.has_value())
		{
			return T(value);
		}
		else
		{
			return {};
		}
	}

	/*!
	    @brief Get the values of multiple datapoints as raw bytes

	    @param dpIds A list of datapoint ids
	    @param reload If true bypass the cache and force a reload from the device
	    @return DatapointValues
	 */
	protocol::DatapointValues getDpValuesRaw(std::vector<std::uint16_t> dpIds, bool reload = false);

	/*!
	    @brief Get the values of multiple datapoints as raw bytes

	    @param startDpId The id of the first datapoint value to request
	    @param endDpId The id of the last datapoint value to request
	    @param reload [opt] If true bypass the cache and force a reload from the device, defaults to false
	    @param filter [opt] The filter to select a subset of datapoint values, defaults to DatapointValueFilter::VALID
	    @return protocol::DatapointValues
	 */
	protocol::DatapointValues getDpValuesRaw(
	    std::uint16_t startDpId,
	    std::uint16_t endDpId,
	    bool reload                           = false,
	    protocol::DatapointValueFilter filter = protocol::DatapointValueFilter::VALID);

	/*!
	    @brief Load all the datapoint values and cache them.

	    To detemine how many serveritems are available the serverItem maxDatapoints is used. If this is not available
	   load 1-65535.
	    @param reload [opt] If true bypass the cache and force a reload from the device, defaults to false
	    @param filter [opt] The filter to select a subset of datapoint values, defaults to DatapointValueFilter::VALID
	    @return DatapointConfigurations
	 */
	protocol::DatapointValues getAllDpValuesRaw(
	    bool reload = false, protocol::DatapointValueFilter filter = protocol::DatapointValueFilter::VALID);

	///@}
	/*************************************** Configuration of this class *****************************/

	/*!
	    @brief Set the strategy to use for reading the serveritems

	    @param strategy The strategy to use
	    @return The previous strategy
	 */
	ServerItemReadStrategy setServerItemReadStrategy(ServerItemReadStrategy strategy);

	/*!
	    @brief Enable or Disable the internal caching for datapoint and parameter bytes.

	    Serveritems are always cached

	    @param state The new state to set
	    @return The previous state is new state was different or the current unchanged state
	 */
	bool enableCaching(bool state);

	/*!
	    @brief Return the current caching state

	    @return true if caching is enabled or false otherwise
	 */
	bool isCachingEnabled() const
	{
		return config_.cacheEnabled;
	};

	/*!
	    @brief Clear the internal caches for datapoints and parameter bytes
	 */
	void clearCaches();

	/*!
	    @brief Returns if the last request could be statisfied by the cache or not

	    @return true request could be answered by cache, false otherwise
	 */
	bool wasCacheHit()
	{
		return cacheHit_;
	}

	/*!
	    @brief Get the metadata set for this device if available

	    @return std::optional<DeviceMeta> The metadata or an empty std::optional
	 */
	std::optional<DeviceMeta> getMetaInfo();

	/*!
	    @brief Set the meta data for this device

	    In most cases this is not necessary. Gathering information from the serveritems an DeviceGroupIdentifier is
	    generated and then a lookup in DevicesMetaLookup is performed. For devices groups not in this lookup table this
	    function may be used to set the metainformation beforehand. If incorrect values are set the SDK wont work as
	    expected.
	    @param meta The meta data for this device
	 */
	void setMetaInfo(DeviceMeta&& meta);


	/*************************************** Callbacks on events *************************************/

	/*!
	    @brief Register one or more functions ascallbacks, invoked on serveritem changes
	*/
	void registerServerItemCallback(ServerItemIndicationCb);

	/*!
	    @brief Register one or more functions ascallbacks, invoked on datapoint value changes
	*/
	void registerDPValueCallback(DatapointValueCb);


	/************************************ Underlying BaosConnection *********************************/

	/*!
	    @brief Pure virtual function to get a BaosConnection to the device

	    @return BaosConnection_UPtr A BaosConnection to the device
	 */
	virtual BaosConnection_UPtr getConnection() = 0;

	/// @cond TESTING
	/*!
	    @brief Set the Internal Connection object used by this class.

	    @remark Only intended for testing
	    @param conn The BaosConnection object which this class should use to communicate with the BAOS device
	 */
	void setInternalConnection(BaosConnection_UPtr& conn)
	{
		baosCon_ = std::move(conn);
	}

	/*!
	    @brief Get the internal baos connection used by this class.

	    @remark Only intended for testing
	    @return a referencte to the BaosConnection
	 */
	BaosConnection& getInternalConnection()
	{
		if (!baosCon_)
		{
			throw WzSocketException(WZ_SL, "No internal connection available");
		}
		return *baosCon_;
	}
	/// @endcond


protected:
	/*!
	    @brief Make sure that a connection to the underlying device exists and is open or throw an exception if this
	    is not possible
	 */
	void ensureConnection();

	/*!
	    @brief An update may occur as a response to a getServerItem request or via an indication

	    @param serverItems The Serveritems that were changed
	 */
	void onServerItemsUpdate(const protocol::ServerItems& serverItems);

	/*!
	    @brief An update may occur as a response to a getDatapointValue request or via an indication

	    @param dpValueStates  The datapoint values and their states that got changed
	 */
	void onDpValueUpdate(const protocol::DatapointValueStates& dpValueStates);

	/*!
	    @brief Reset the device information, retrieved from serveritems and meta data to their default values

	 */
	void clearDeviceConfigs();

	/*!
	    @brief Get all the serveritem values from the device.

	    If the wz::baos::ServerItemReadStrategy is set to ServerItemReadStrategy::ALLATONCE request as many serveritems
	   from the device as it possible. If the ServerItemReadStrategy is set to ServerItemReadStrategy::METADATAREAD,
	   first check if meta information is already set, e.g. by calling setMetaInfo. If the meta information is available
	   use it to determine what the hightest server item id is and request all the serveritem including it. On the other
	   hand if the meta information is not available, read the serveritems [ManufactureCodeApp, ApplicationId,
	   ApplicationVersion] generate a DeviceGroupIdentifier from it and check if DevicesMetaLookup has an entry for it.
	   If no entry is found throw an exception when ServerItemReadStrategy::METADATAREAD is set or fallback to ALLATONCE
	   if METADATAREADWITHFALLBACK was configured as serveritem read strategy. If found use the maxServerItemId meta
	   information to load the serveritems valus from id [1;maxServerItemId]

	    @return A list of ServerItem objects
	 */
	protocol::ServerItems getAllServerItems();

	/*!
	    @brief Helper function to retrieve the elements(datapoint descriptions, string and parameter bytes) from the
	   device

	    @tparam TGetter The SDK function type used to get a chunk of data from the device
	    @param start The id of the first element
	    @param count The last element is then id+count-1
	    @param maxLengthElement The maximum of bytes one element may contain
	    @param cb The callback invoked for every rquest to the BAOS device
	    @return The reponse as a list of bytes
	 */
	template<class TGetter>
	auto getElements(std::uint16_t start, std::uint16_t count, std::uint16_t maxLengthElement, TGetter cb)
	{
		if (config_.blockReadEnabled)
		{
			// Get current buffer size
			if (!deviceInfoLoaded_)
			{
				getConfigInfo();
			}
			assert(configInfo_.currentBufferSize > 16);
			std::uint16_t payloadAvailSize = (configInfo_.currentBufferSize - 16) /
			                                 maxLengthElement;  // 10 Byte knxip header and 6 bytes BAOS header and
			                                                    // maxLengthElement bytes per payload element
			assert(payloadAvailSize > 0);
			return utility::BlockReader::read(start, count, payloadAvailSize, cb);
		}
		else
		{
			return cb(start, count);
		}
	}


protected:
	DeviceClassConfig config_{
	    true, true, ServerItemReadStrategy::METADATAREADWITHFALLBACK};  //!< Configuration options for this class.
	std::optional<DeviceMeta> metaInfo_;                                //!< Meta information set by the user

	bool cacheHit_{false};  //!< Flag to indicate that the last operation could be statisfied with the cache content

	// cached variables
	bool deviceInfoLoaded_{false};   //!< Flag set of the device information is loaded from the device via serveritems
	DeviceStaticInfo staticInfo_;    //!< Static information about the BAOS device
	DeviceConfigInfo configInfo_;    //!< Config information about the BAOS device
	DeviceRuntimeInfo runtimeInfo_;  //!< Runtime information about the BAOS device

	std::map<std::uint16_t, Buffer> dpCache_;  //!< Cache to hold the last valid values for the datapoints
	bool dpCacheFilled_{false};                //!< Flag to indicate that the cache is valid
	protocol::ParameterBytes paramBytes_;      //!< Cache to hold the parameter bytes

	DatapointConfigurations dpConfig_;  //!< Configuration information about the BAOS datapoints

	std::vector<ServerItemIndicationCb> serverItemsCallbacks;  //!< List of registered callback functions
	std::vector<DatapointValueCb> dpValueCallbacks;            //!< List of registered callback functions

	BaosConnection_UPtr baosCon_;  //!< The internal connection used to coomunicate with the BAOS device

	WZLOGGER("BaosDevice", LVL_DEBUG);  //!< The name for the logger of this class
};

}  // namespace wz::baos

#endif  // __BAOSLIB_INTERFACE_H__
