//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//
#ifndef __BAOSLIB_INTERFACE_META_H__
#define __BAOSLIB_INTERFACE_META_H__

#include "baos/protocol/Defines.h"

#include "wzcpp/features/utils/Formatter.h"


/// @cond DOCUMENT_STL
#include <array>
#include <cassert>
#include <cstdint>
#include <functional>
#include <map>
#include <string>
/// @endcond

/*!
    @file
    @brief  Provides the structures for additional meta information regarding the various devices
 */

namespace wz::baos
{
/*!
    @brief Unique identifier for a device

    The combination of manufacturer code of the ETS app, the application id and version allows to target a specific
   group of devices
 */
struct DeviceGroupIdentifier
{
	/*!
	    @brief Construct a new Device Identifier object build from baos::protocol::ServerItem elements

	    @param manufacturerApp The manufacturer code of the ETS App
	    @param applicationId  The application id of the ETS App
	    @param applicationVersion The application version of the ETS App
	 */
	DeviceGroupIdentifier(const protocol::ServerItem& manufacturerApp,
	                      const protocol::ServerItem& applicationId,
	                      const protocol::ServerItem& applicationVersion)
	{
		assert(manufacturerApp.id == protocol::ServerItemId::ManufactureCodeApp);
		assert(applicationId.id == protocol::ServerItemId::ApplicationId);
		assert(applicationVersion.id == protocol::ServerItemId::ApplicationVersion);

		deviceid[0] = static_cast<std::byte>(manufacturerApp.data[0] & 0xff);
		deviceid[1] = static_cast<std::byte>(manufacturerApp.data[1] & 0xff);
		deviceid[2] = static_cast<std::byte>(applicationId.data[0] & 0xff);
		deviceid[3] = static_cast<std::byte>(applicationId.data[1] & 0xff);
		deviceid[4] = static_cast<std::byte>(applicationVersion.data[0] & 0xff);
	};

	/*!
	    @brief Construct a new Device Identifier object from intergral values

	    @param manufacturerApp The manufacturer code of the ETS App
	    @param applicationId  The application id of the ETS App
	    @param applicationVersion The application version of the ETS App
	 */
	DeviceGroupIdentifier(std::uint16_t manufacturerApp, std::uint16_t applicationId, std::uint8_t applicationVersion)
	{
		deviceid[0] = static_cast<std::byte>((manufacturerApp >> 8) & 0xff);
		deviceid[1] = static_cast<std::byte>(manufacturerApp & 0xff);
		deviceid[2] = static_cast<std::byte>((applicationId >> 8) & 0xff);
		deviceid[3] = static_cast<std::byte>(applicationId & 0xff);
		deviceid[4] = static_cast<std::byte>(applicationVersion & 0xff);
	};

	//! The comparision operator for the class
	bool operator==(const DeviceGroupIdentifier& other) const
	{
		return (deviceid == other.deviceid);
	}
	//! The lesser than operator of this class
	bool operator<(const DeviceGroupIdentifier& other) const
	{
		return (deviceid < other.deviceid);
	}

	std::array<std::byte, 5> deviceid;  //!< The unique inferfacegroup id
};

/// Meta information about a device group, which is required be the device class
struct DeviceMeta
{
	using UPtr = std::unique_ptr<DeviceMeta>;  //!< An alias for an unique pointer to this class
	std::string name;                          //!< The name of the devices targeted with this metainformation block
	std::uint16_t maxServerItemId{0};          //!< The maximum amount if serveritems in this devices
	std::uint16_t maxParameterBytes{0};        //!< The maximum of parameter bytes in this devices
	std::uint16_t maxDatapoints{0};            //!< The maximum of datapoints in this deivces
};

using DevicesMeta =
    std::map<DeviceGroupIdentifier, DeviceMeta>;  //!< Maps the unqiue device identifier to an metainformation entry

/*!
    @brief Global lookup map for known devices( via DeviceGroupIdentifier as key) and their metadata as value
*/
static DevicesMeta DevicesMetaLookup = {
    {{0xC5, 0x0708, 0x10}, DeviceMeta{"771", 18, 250, 250}},
    {{0xC5, 0x0709, 0x10}, DeviceMeta{"772", 18, 250, 1000}},  // despite having 1000 parameter bytes configurable in
                                                               // the ETS, the service only allows access to 250
    {{0xC5, 0x0714, 0x10}, DeviceMeta{"773", 18, 250, 250}},
    {{0xC5, 0x0715, 0x10}, DeviceMeta{"774", 18, 1000, 1000}}};
}  // namespace wz::baos

namespace std
{

/*!
    @brief Custom hash struct for DeviceGroupIdentifier

    Tho allow the usage of DeviceGroupIdentifier as key in the STL cointainers like map, a specialization of
    the hash struct for this type must be provided.
    This simple version XORs the parts of the 5 byte array to get a hexadecimal string which then is hashed.
 */
template<>
struct hash<wz::baos::DeviceGroupIdentifier>
{
	/*!
	@brief The overloaded call operator returning a hash value

	@param k The deviceidentifier to hash
	@return std::size_t The hashed value
	 */
	std::size_t operator()(const wz::baos::DeviceGroupIdentifier& k) const
	{
		using std::hash;
		using std::size_t;
		using std::string;

		auto hexString = wzcpp::rangeToHex<std::array<std::byte, 5>>(k.deviceid);
		return std::hash<std::string>()(hexString);
	}
};

}  // namespace std

#endif  // end ifdef __BAOSLIB_INTERFACE_META_H__
