//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#ifndef __BAOSLIB_IPV4INTERFACE_H__
#define __BAOSLIB_IPV4INTERFACE_H__


#include "baos/config.h"
#include "baos/device/BaosDevice.h"
#include "baos/protocol/Defines.h"

/// @cond DOCUMENT_STL
#include <memory>
#include <optional>
#include <utility>
/// @endcond


/*!
    @file
    @brief Specialization for a IP v4 BAOSDevice
*/

namespace wz::baos::ip
{


/*******************************************
 * BaosIp4Device
 *******************************************/

/*!
    @brief  Represents a BAOS IP4 device with an underlying TCP connection

    The connection information like ip address and the port can either be set in the constructor or later with
   setConnectionData. This has to be done before opening a the connection to the device.
 */
class BAOSLIB_EXPORT BaosIp4Device : public wz::baos::BaosDevice
{
public:
	virtual ~BaosIp4Device() = default;

	/*!
	    @brief Construct a new Baos Ip 4 Device object from an ip address and a port

	    @param ip4Address  The IPv4 Address as string in the form "xxx.xxx.xxx.xxx"
	    @param port The tcp port of the remote device
	 */
	BaosIp4Device(const std::string& ip4Address, std::uint16_t port = 12004)
	    : ip4Address_{ip4Address},
	      port_{port} {

	      };

	/*!
	    @brief Construct a new Baos Ip 4 Device object from an ip address and a port

	    @param ip4Address  The IPv4 Address as string in the form "xxx.xxx.xxx.xxx"
	    @param port The tcp port of the remote device

	    @remark The connection information is NOT automatically used. To apply it a manual call to close is required.
	    The connection will then be reopened on the next request.
	 */
	void setConnectionData(const std::string& ip4Address, std::uint16_t port = 12004)
	{
		ip4Address_ = ip4Address;
		port_       = port;
		clearCaches();
		clearDeviceConfigs();
		deviceInfoLoaded_ = false;
	}

	/*!
	    @brief Get a BaosConnection to the device

	    @return BaosConnection_UPtr A BaosConnection to the device
	 */
	BaosConnection_UPtr getConnection() override;


private:
	std::string ip4Address_;
	std::uint16_t port_;
	std::unique_ptr<Ip4Driver> driver_;
	WZLOGGER("BaosIp4Device", LVL_DEBUG);
};

}  // namespace wz::baos::ip

#endif  // __BAOSLIB_IPV4INTERFACE_H__
