//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#ifndef __BAOSLIB_IND_SENDER_H__
#define __BAOSLIB_IND_SENDER_H__

#include "baos/BaosLib_export.h"
#include "baos/config.h"
#include "baos/protocol/Indication.h"


#include <array>
#include <atomic>
#include <memory>
#include <string>


namespace wz::baos
{


/******************************************************************
 * BaosIndicationSender
 ******************************************************************/

/*!
    @brief Receiving indications from the framework, this class runs in its own thread and invokes the registered
   callback for every indication
 */
class BAOSLIB_EXPORT BaosIndicationSender
{
public:
	using UPtr                   = std::unique_ptr<BaosIndicationSender>;
	using UniqueLock             = std::unique_lock<std::mutex>;
	using ServerItemIndicationCb = std::function<void(const protocol::ServerItems&)>;
	using DatapointValueCb       = std::function<void(const protocol::DatapointValueStates&)>;

	BaosIndicationSender(protocol::BaosIndQueue& indQueue);
	~BaosIndicationSender();

	void run(); /*!< Starts the mainloop for this class */
	void init() /*!< Initialises the internal state */
	{
		shouldStop_ = false;
	}
	void stop();            /*!< Request the mainloop to stop */
	void waitForFinished(); /*!< Wait until the mainloop has finished*/
	bool isRunning()        /*!< Status of the mainloop */
	{
		return !stopped_;
	};

	void setServerItemsIndCallback(ServerItemIndicationCb callback)
	{
		std::scoped_lock<std::mutex> lk{dataMut_};
		serverItemIndCallback_ = callback;
	}

	void setDatapointValueIndCallback(DatapointValueCb callback)
	{
		std::scoped_lock<std::mutex> lk{dataMut_};
		datapointValueIndCallback_ = callback;
	}

protected:
	static std::uint64_t indId; /*!< A continious id for each indication event */

	std::atomic<bool> shouldStop_{false}; /*!< A flag signaling the main loop to stop */

	std::condition_variable cvFinished_; /*!< This condition signals if the main loop has acutally stoppped */
	mutable std::mutex cvFinished_mut_;  /*!< The mutext related to the condition variable cvFinished_ */
	std::atomic<bool> stopped_{false};   /*!< Flag that shows the actual stoppped state*/

	protocol::BaosIndQueue& indQueue_;             /*!< A concurrent queue for the indications to send */
	mutable std::mutex dataMut_;                   /*!< The mutext protecting the callback function objects*/
	ServerItemIndicationCb serverItemIndCallback_; /*!< Stores the callback to invoke on serveritem indications */
	DatapointValueCb datapointValueIndCallback_;   /*!< Stores the callback to invoke on datapoint indications */
};

}  // namespace wz::baos

#endif  //__BAOSLIB_IND_SENDER_H__
