//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#include "Helper.h"

#include <baos/Sdk.h>
#include <baos/ip/BaosTcpConnection.h>
#include <baos/protocol/Defines.h>
#include <wzcpp/features/utils.h>


#include <iostream>
#include <memory>

/*!
    @example GetDatapointDescription.cpp

    This sample shows how to read datapoint descriptions via the connection API <br>

    Usage:
    @code
        GetDatapointDescription <ip_address>
        // e.g.:
        GetDatapointDescription 10.0.0.102
    @endcode

    @note Another possibility is via the device API. <br>
    See also: Sample 01_BaosDatapointConfiguration.cpp
*/

using utility::CommandLineOptions;
using wz::baos::setBaosLogLevel;
using namespace wz::baos;
using namespace wz::baos::ip;
using namespace wz::baos::protocol;

/******************************************
** Main
*******************************************/

int main(int argc, char* argv[])
{
	try
	{
		std::cout << "******************************************" << std::endl;
		std::cout << "GetDatapointDescription sample" << std::endl;
		std::cout << "******************************************" << std::endl;
		std::cout << std::endl;

		// Reduce log messages
		setBaosLogLevel(wzcpp::LogLevel::warn);

		// The CommandLineOptions::parse function
		// validates the arguments and handles the help command
		CommandLineOptions options;
		options.parse(argc, argv);
		if (options.wasHelpdisplayed())
		{
			return 0;
		}

		// Get the ip address from the parsed command line options
		const std::string ipAddress = options.getIpAddress();
		std::cout << " Connection requested for " << ipAddress << std::endl;

		// Create a TPC/IP connection to the remote BAOS device
		BaosTcpConnection connection;
		connection.open(ipAddress);

		// Reads some datapoint descriptions from the device (single item per requested)

		// Read datapoint description of id 1
		std::cout << "Reading description of dp id 1 ... " << std::endl;
		auto dpDescription = connection.getDpDescription(1);
		if (dpDescription.has_value())
		{
			std::cout << "DP size: " << dpSizeToString(dpDescription->size) << std::endl;
			std::cout << "DP flags: " << dpDescription->flags.toString() << std::endl;
			std::cout << "DPT: " << dptToString(dpDescription->dpt) << std::endl;
			std::cout << std::endl;
		}

		// Read datapoint description of id == 10
		std::cout << "Reading dp description of id == 10 ... " << std::endl;
		dpDescription = connection.getDpDescription(10);
		if (dpDescription.has_value())
		{
			std::cout << "DP size: " << dpSizeToString(dpDescription->size) << std::endl;
			std::cout << "DP flags: " << dpDescription->flags.toString() << std::endl;
			std::cout << "DPT: " << dptToString(dpDescription->dpt) << std::endl;
			std::cout << std::endl;
		}

		// Reads multiple datapoint descriptions (multiple items per requested)

		// start = 1; number of = 1000
		std::cout << "Reading dp descriptions of id == 1 to 1000 ... " << std::endl;
		const DatapointDescriptions descriptions = connection.getDpDescriptions(1, 1000);

		// trace the descriptions
		for (const auto& item : descriptions)
		{
			std::cout << "Datapoint id " << item.first << std::endl;

			const DatapointDescription& desc = item.second;
			std::cout << "DP size: " << dpSizeToString(desc.size) << std::endl;
			std::cout << "DP flags: " << desc.flags.toString() << std::endl;
			std::cout << "DPT: " << dptToString(desc.dpt) << std::endl;
			std::cout << std::endl;
		}
	}
	catch (const std::exception& e)
	{
		std::cerr << "Failed: " << e.what() << std::endl;
		return -1;
	}

	return 0;
}
