//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#include "Helper.h"

#include <baos/Sdk.h>
#include <baos/ip/BaosTcpConnection.h>
#include <baos/protocol/Defines.h>
#include <wzcpp/features/utils.h>

#include <iostream>
#include <memory>

/*!
    @example GetDatapointValue.cpp

    This sample shows how to get datapoint values via the connection API <br>


    Datapoint

    Usage:
    @code
        GetDatapointValue <ip_address>
        // e.g.:
        GetDatapointValue 10.0.0.102
    @endcode

    Pre-condition:
    It assumes that the following datapoints are configured:

    @verbatim

    |---------|-----------------|-----------------------------|-------------|---------------|
    | Number  |  DPT            |  Name                       |  SizeInBit  |  Flags        |
    |         |                 |                             |             | C R W T U I   |
    |---------|-----------------|-----------------------------|-------------|---------------|
    |    1    | DPT 01 - 1 bit  | Actuator A - Switch         | 1           | C - W - - -   |
    |---------|-----------------|-----------------------------|-------------|---------------|
    |    2    | DPT 01 - 1 bit  | Actuator A - State          | 1           | C R - T - -   |
    |---------|-----------------|-----------------------------|-------------|---------------|
    |    3    | DPT 01 - 1 bit  | Actuator B - Switch         | 1           | C - W - - -   |
    |---------|-----------------|-----------------------------|-------------|---------------|
    |    4    | DPT 01 - 1 bit  | Actuator B - State          | 1           | C R - T - -   |
    |---------|-----------------|-----------------------------|-------------|---------------|
    @endverbatim

    @note Another possibility is via the device API. <br>
    See also: Sample 04_BaosDatapointValues.cpp
*/

using utility::CommandLineOptions;
using wz::baos::setBaosLogLevel;
using namespace wz::baos;
using namespace wz::baos::ip;
using namespace wz::baos::protocol;

/******************************************
** Anonymous namespace
*******************************************/

namespace
{

/*!
    Trace the parameter bytes
*/
void traceDatapointValue(const DatapointValue& dpValue)
{
	std::cout.fill('0');
	std::cout << std::hex << std::uppercase;
	for (auto& item : dpValue)
	{
		std::cout << std::setw(2) << static_cast<int>(item) << " ";
	}
	std::cout << std::nouppercase << std::dec;
}

}  // end anonymous namespace

/******************************************
** Main
*******************************************/

int main(int argc, char* argv[])
{
	try
	{
		std::cout << "******************************************" << std::endl;
		std::cout << "GetDatapointValue sample" << std::endl;
		std::cout << "******************************************" << std::endl;
		std::cout << std::endl;

		// Reduce log messages
		setBaosLogLevel(wzcpp::LogLevel::warn);

		// The CommandLineOptions::parse function
		// validates the arguments and handles the help command
		CommandLineOptions options;
		options.parse(argc, argv);
		if (options.wasHelpdisplayed())
		{
			return 0;
		}

		// Get the ip address from the parsed command line options
		const std::string ipAddress = options.getIpAddress();
		std::cout << " Connection requested for " << ipAddress << std::endl;

		// Create a TPC/IP connection to the remote BAOS device
		BaosTcpConnection connection;
		connection.open(ipAddress);


		// Reads some datapoint values (single item per requested)

		// id = 2 (DPT 01 - 1 bit)
		std::cout << "Reading datapoint value id 2:" << std::endl;
		auto valueState = connection.getDpValue(2);
		if (valueState.has_value())
		{
			std::cout << "Id: 2, value: ";
			traceDatapointValue(valueState->value);
			std::cout << std::endl;
		}

		// id = 4 (DPT 01 - 1 bit)
		std::cout << "Reading datapoint value id 4:" << std::endl;
		valueState = connection.getDpValue(4);
		if (valueState.has_value())
		{
			std::cout << "Id 4, value: ";
			traceDatapointValue(valueState->value);
			std::cout << std::endl;
		}

		// Reads multiple datapoint values (multiple items per requested)
		// Return only the valid datapoints

		// id = 1; numbers = 100
		std::cout << "Reading datapoint value ids 1 - 100:" << std::endl;
		DatapointValueStates valueStates = connection.getDpValues(1, 100, DatapointValueFilter::ALL);

		// print datapoint id and it's data buffer as hex string
		for (const auto& item : valueStates)
		{
			const DatapointValueState& value = item.second;
			std::cout << "Id: " << item.first;
			std::cout << ", value: ";
			traceDatapointValue(value.value);
			std::cout << std::endl;
		}
	}
	catch (const std::exception& e)
	{
		std::cerr << "Failed: " << e.what() << std::endl;
		return -1;
	}

	return 0;
}
